@@ -32,6 +32,7 @@
 #include "radv_entrypoints.h"
 #include "radv_image_view.h"
 #include "radv_physical_device.h"
+#include "radv_query.h"
 #include "radv_video.h"
 
 #include "ac_vcn_enc.h"
@@ -112,21 +113,44 @@ void
 radv_probe_video_encode(struct radv_physical_device *pdev)
 {
    pdev->video_encode_enabled = false;
+
+   /* TODO: Add VCN 5.0+. */
+   if (pdev->info.vcn_ip_version >= VCN_5_0_0)
+      return;
+
    if (pdev->info.vcn_ip_version >= VCN_4_0_0) {
       if (pdev->info.vcn_enc_major_version != RENCODE_V4_FW_INTERFACE_MAJOR_VERSION)
          return;
       if (pdev->info.vcn_enc_minor_version < RENCODE_V4_FW_INTERFACE_MINOR_VERSION)
          return;
+
+      /* VCN 4 FW 1.22 has all the necessary pieces to pass CTS */
+      if (pdev->info.vcn_enc_minor_version >= 22) {
+         pdev->video_encode_enabled = true;
+         return;
+      }
    } else if (pdev->info.vcn_ip_version >= VCN_3_0_0) {
       if (pdev->info.vcn_enc_major_version != RENCODE_V3_FW_INTERFACE_MAJOR_VERSION)
          return;
       if (pdev->info.vcn_enc_minor_version < RENCODE_V3_FW_INTERFACE_MINOR_VERSION)
          return;
+
+      /* VCN 3 FW 1.33 has all the necessary pieces to pass CTS */
+      if (pdev->info.vcn_enc_minor_version >= 33) {
+         pdev->video_encode_enabled = true;
+         return;
+      }
    } else if (pdev->info.vcn_ip_version >= VCN_2_0_0) {
       if (pdev->info.vcn_enc_major_version != RENCODE_V2_FW_INTERFACE_MAJOR_VERSION)
          return;
       if (pdev->info.vcn_enc_minor_version < RENCODE_V2_FW_INTERFACE_MINOR_VERSION)
          return;
+
+      /* VCN 2 FW 1.24 has all the necessary pieces to pass CTS */
+      if (pdev->info.vcn_enc_minor_version >= 24) {
+         pdev->video_encode_enabled = true;
+         return;
+      }
    } else {
       if (pdev->info.vcn_enc_major_version != RENCODE_FW_INTERFACE_MAJOR_VERSION)
          return;
@@ -451,15 +475,16 @@ radv_enc_session_init(struct radv_cmd_buffer *cmd_buffer, const struct VkVideoEn
    const struct radv_physical_device *pdev = radv_device_physical(device);
    struct radv_video_session *vid = cmd_buffer->video.vid;
    struct radeon_cmdbuf *cs = cmd_buffer->cs;
-   unsigned alignment = 16;
+   unsigned alignment_w = 16;
+   unsigned alignment_h = 16;
    if (vid->vk.op == VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR) {
-      alignment = 64;
+      alignment_w = 64;
    }
 
    uint32_t w = enc_info->srcPictureResource.codedExtent.width;
    uint32_t h = enc_info->srcPictureResource.codedExtent.height;
-   uint32_t aligned_picture_width = align(w, alignment);
-   uint32_t aligned_picture_height = align(h, alignment);
+   uint32_t aligned_picture_width = align(w, alignment_w);
+   uint32_t aligned_picture_height = align(h, alignment_h);
    uint32_t padding_width = aligned_picture_width - w;
    uint32_t padding_height = aligned_picture_height - h;
 
@@ -515,8 +540,8 @@ radv_enc_slice_control(struct radv_cmd_buffer *cmd_buffer, const struct VkVideoE
    struct radeon_cmdbuf *cs = cmd_buffer->cs;
 
    uint32_t num_mbs_in_slice;
-   uint32_t width_in_mbs = enc_info->srcPictureResource.codedExtent.width / 16;
-   uint32_t height_in_mbs = enc_info->srcPictureResource.codedExtent.height / 16;
+   uint32_t width_in_mbs = DIV_ROUND_UP(enc_info->srcPictureResource.codedExtent.width, 16);
+   uint32_t height_in_mbs = DIV_ROUND_UP(enc_info->srcPictureResource.codedExtent.height, 16);
    num_mbs_in_slice = width_in_mbs * height_in_mbs;
    ENC_BEGIN;
    radeon_emit(cs, pdev->vcn_enc_cmds.slice_control_h264);
@@ -593,16 +618,11 @@ radv_enc_slice_control_hevc(struct radv_cmd_buffer *cmd_buffer, const struct VkV
    struct radv_device *device = radv_cmd_buffer_device(cmd_buffer);
    const struct radv_physical_device *pdev = radv_device_physical(device);
    struct radeon_cmdbuf *cs = cmd_buffer->cs;
-   const struct VkVideoEncodeH265PictureInfoKHR *h265_picture_info =
-      vk_find_struct_const(enc_info->pNext, VIDEO_ENCODE_H265_PICTURE_INFO_KHR);
-   const StdVideoEncodeH265PictureInfo *pic = h265_picture_info->pStdPictureInfo;
-   const StdVideoH265SequenceParameterSet *sps =
-      vk_video_find_h265_enc_std_sps(&cmd_buffer->video.params->vk, pic->pps_seq_parameter_set_id);
 
    uint32_t width_in_ctb, height_in_ctb, num_ctbs_in_slice;
 
-   width_in_ctb = sps->pic_width_in_luma_samples / 64;
-   height_in_ctb = sps->pic_height_in_luma_samples / 64;
+   width_in_ctb = DIV_ROUND_UP(enc_info->srcPictureResource.codedExtent.width, 64);
+   height_in_ctb = DIV_ROUND_UP(enc_info->srcPictureResource.codedExtent.height, 64);
    num_ctbs_in_slice = width_in_ctb * height_in_ctb;
    ENC_BEGIN;
    radeon_emit(cs, pdev->vcn_enc_cmds.slice_control_hevc);
@@ -1245,7 +1265,7 @@ radv_enc_bitstream(struct radv_cmd_buffer *cmd_buffer, struct radv_buffer *buffe
    struct radv_device *device = radv_cmd_buffer_device(cmd_buffer);
    const struct radv_physical_device *pdev = radv_device_physical(device);
    struct radeon_cmdbuf *cs = cmd_buffer->cs;
-   uint64_t va = radv_buffer_get_va(buffer->bo) + buffer->offset;
+   uint64_t va = radv_buffer_get_va(buffer->bo) + buffer->offset + offset;
    radv_cs_add_buffer(device->ws, cs, buffer->bo);
 
    ENC_BEGIN;
@@ -1254,12 +1274,12 @@ radv_enc_bitstream(struct radv_cmd_buffer *cmd_buffer, struct radv_buffer *buffe
    radeon_emit(cs, va >> 32);
    radeon_emit(cs, va & 0xffffffff);
    radeon_emit(cs, buffer->vk.size);
-   radeon_emit(cs, offset);
+   radeon_emit(cs, 0);
    ENC_END;
 }
 
 static void
-radv_enc_feedback(struct radv_cmd_buffer *cmd_buffer)
+radv_enc_feedback(struct radv_cmd_buffer *cmd_buffer, uint64_t feedback_query_va)
 {
    struct radv_device *device = radv_cmd_buffer_device(cmd_buffer);
    const struct radv_physical_device *pdev = radv_device_physical(device);
@@ -1268,8 +1288,8 @@ radv_enc_feedback(struct radv_cmd_buffer *cmd_buffer)
 
    radeon_emit(cs, pdev->vcn_enc_cmds.feedback);
    radeon_emit(cs, RENCODE_FEEDBACK_BUFFER_MODE_LINEAR);
-   radeon_emit(cs, cmd_buffer->video.feedback_query_va >> 32);
-   radeon_emit(cs, cmd_buffer->video.feedback_query_va & 0xffffffff);
+   radeon_emit(cs, feedback_query_va >> 32);
+   radeon_emit(cs, feedback_query_va & 0xffffffff);
    radeon_emit(cs, 16); // buffer_size
    radeon_emit(cs, 40); // data_size
    ENC_END;
@@ -1360,6 +1380,7 @@ radv_enc_params(struct radv_cmd_buffer *cmd_buffer, const VkVideoEncodeInfoKHR *
    uint64_t chroma_va = va + src_img->planes[1].surface.u.gfx9.surf_offset;
    uint32_t pic_type;
    unsigned int slot_idx = 0xffffffff;
+   unsigned int max_layers = cmd_buffer->video.vid->rc_layer_control.max_num_temporal_layers;
 
    radv_cs_add_buffer(device->ws, cs, src_img->bindings[0].bo);
    if (h264_pic) {
@@ -1378,7 +1399,7 @@ radv_enc_params(struct radv_cmd_buffer *cmd_buffer, const VkVideoEncodeInfoKHR *
          pic_type = RENCODE_PICTURE_TYPE_I;
          break;
       }
-      radv_enc_layer_select(cmd_buffer, h264_pic->temporal_id);
+      radv_enc_layer_select(cmd_buffer, MIN2(h264_pic->temporal_id, max_layers));
    } else if (h265_pic) {
       switch (h265_pic->pic_type) {
       case STD_VIDEO_H265_PICTURE_TYPE_P:
@@ -1395,7 +1416,7 @@ radv_enc_params(struct radv_cmd_buffer *cmd_buffer, const VkVideoEncodeInfoKHR *
          pic_type = RENCODE_PICTURE_TYPE_I;
          break;
       }
-      radv_enc_layer_select(cmd_buffer, h265_pic->TemporalId);
+      radv_enc_layer_select(cmd_buffer, MIN2(h265_pic->TemporalId, max_layers));
    } else {
       assert(0);
       return;
@@ -1467,15 +1488,6 @@ radv_enc_op_init(struct radv_cmd_buffer *cmd_buffer)
    ENC_END;
 }
 
-static void
-radv_enc_op_close(struct radv_cmd_buffer *cmd_buffer)
-{
-   struct radeon_cmdbuf *cs = cmd_buffer->cs;
-   ENC_BEGIN;
-   radeon_emit(cs, RENCODE_IB_OP_CLOSE_SESSION);
-   ENC_END;
-}
-
 static void
 radv_enc_op_enc(struct radv_cmd_buffer *cmd_buffer)
 {
@@ -1510,6 +1522,13 @@ radv_enc_op_preset(struct radv_cmd_buffer *cmd_buffer, const VkVideoEncodeInfoKH
    struct radv_video_session *vid = cmd_buffer->video.vid;
    uint32_t preset_mode;
 
+   if (vid->enc_preset_mode == RENCODE_PRESET_MODE_QUALITY)
+      preset_mode = RENCODE_IB_OP_SET_QUALITY_ENCODING_MODE;
+   else if (vid->enc_preset_mode == RENCODE_PRESET_MODE_BALANCE)
+      preset_mode = RENCODE_IB_OP_SET_BALANCE_ENCODING_MODE;
+   else
+      preset_mode = RENCODE_IB_OP_SET_SPEED_ENCODING_MODE;
+
    switch (vid->vk.op) {
    case VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR: {
       const struct VkVideoEncodeH265PictureInfoKHR *h265_picture_info =
@@ -1517,22 +1536,14 @@ radv_enc_op_preset(struct radv_cmd_buffer *cmd_buffer, const VkVideoEncodeInfoKH
       const StdVideoEncodeH265PictureInfo *pic = h265_picture_info->pStdPictureInfo;
       const StdVideoH265SequenceParameterSet *sps =
          vk_video_find_h265_enc_std_sps(&cmd_buffer->video.params->vk, pic->pps_seq_parameter_set_id);
-      if (sps->flags.sample_adaptive_offset_enabled_flag && vid->enc_preset_mode == RENCODE_PRESET_MODE_SPEED) {
+      if (sps->flags.sample_adaptive_offset_enabled_flag && vid->enc_preset_mode == RENCODE_PRESET_MODE_SPEED)
          preset_mode = RENCODE_IB_OP_SET_BALANCE_ENCODING_MODE;
-         return;
-      }
       break;
    }
    default:
       break;
    }
 
-   if (vid->enc_preset_mode == RENCODE_PRESET_MODE_QUALITY)
-      preset_mode = RENCODE_IB_OP_SET_QUALITY_ENCODING_MODE;
-   else if (vid->enc_preset_mode == RENCODE_PRESET_MODE_BALANCE)
-      preset_mode = RENCODE_IB_OP_SET_BALANCE_ENCODING_MODE;
-   else
-      preset_mode = RENCODE_IB_OP_SET_SPEED_ENCODING_MODE;
    ENC_BEGIN;
    radeon_emit(cs, preset_mode);
    ENC_END;
@@ -1625,12 +1636,8 @@ radv_enc_headers_hevc(struct radv_cmd_buffer *cmd_buffer, const VkVideoEncodeInf
 static void
 begin(struct radv_cmd_buffer *cmd_buffer, const VkVideoEncodeInfoKHR *enc_info)
 {
-   struct radv_enc_state *enc = &cmd_buffer->video.enc;
    struct radv_video_session *vid = cmd_buffer->video.vid;
 
-   radv_enc_session_info(cmd_buffer);
-   cmd_buffer->video.enc.total_task_size = 0;
-   radv_enc_task_info(cmd_buffer, false);
    radv_enc_op_init(cmd_buffer);
    radv_enc_session_init(cmd_buffer, enc_info);
    if (vid->vk.op == VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR) {
@@ -1655,18 +1662,6 @@ begin(struct radv_cmd_buffer *cmd_buffer, const VkVideoEncodeInfoKHR *enc_info)
    } while (++i < vid->rc_layer_control.num_temporal_layers);
    radv_enc_op_init_rc(cmd_buffer);
    radv_enc_op_init_rc_vbv(cmd_buffer);
-   radeon_emit_direct(cmd_buffer->cs, enc->task_size_offset, enc->total_task_size);
-}
-
-static void
-destroy(struct radv_cmd_buffer *cmd_buffer)
-{
-   struct radv_enc_state *enc = &cmd_buffer->video.enc;
-   radv_enc_session_info(cmd_buffer);
-   cmd_buffer->video.enc.total_task_size = 0;
-   radv_enc_task_info(cmd_buffer, false);
-   radv_enc_op_close(cmd_buffer);
-   radeon_emit_direct(cmd_buffer->cs, enc->task_size_offset, enc->total_task_size);
 }
 
 static void
@@ -1677,7 +1672,7 @@ radv_vcn_encode_video(struct radv_cmd_buffer *cmd_buffer, const VkVideoEncodeInf
    struct radv_device *device = radv_cmd_buffer_device(cmd_buffer);
    const struct radv_physical_device *pdev = radv_device_physical(device);
    struct radv_enc_state *enc = &cmd_buffer->video.enc;
-
+   uint64_t feedback_query_va;
    switch (vid->vk.op) {
    case VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR:
    case VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR:
@@ -1687,29 +1682,50 @@ radv_vcn_encode_video(struct radv_cmd_buffer *cmd_buffer, const VkVideoEncodeInf
       return;
    }
 
-   if (vid->enc_need_begin) {
-      begin(cmd_buffer, enc_info);
-      vid->enc_need_begin = false;
+   radeon_check_space(device->ws, cmd_buffer->cs, 1024);
+
+   if (pdev->enc_hw_ver >= RADV_VIDEO_ENC_HW_4)
+      radv_vcn_sq_header(cmd_buffer->cs, &cmd_buffer->video.sq, RADEON_VCN_ENGINE_TYPE_ENCODE, false);
+
+   const struct VkVideoInlineQueryInfoKHR *inline_queries = NULL;
+   if (vid->vk.flags & VK_VIDEO_SESSION_CREATE_INLINE_QUERIES_BIT_KHR) {
+      inline_queries = vk_find_struct_const(enc_info->pNext, VIDEO_INLINE_QUERY_INFO_KHR);
+
+      if (inline_queries) {
+         VK_FROM_HANDLE(radv_query_pool, pool, inline_queries->queryPool);
+         feedback_query_va = radv_buffer_get_va(pool->bo);
+         feedback_query_va += pool->stride * inline_queries->firstQuery;
+      }
    }
+
+   if (!inline_queries)
+      feedback_query_va = cmd_buffer->video.feedback_query_va;
+
    // before encode
    // session info
    radv_enc_session_info(cmd_buffer);
 
    cmd_buffer->video.enc.total_task_size = 0;
+
    // task info
    radv_enc_task_info(cmd_buffer, true);
 
-   // temporal layers init
-   unsigned i = 0;
-   do {
-      if (vid->enc_need_rate_control) {
+   if (vid->enc_need_begin) {
+      begin(cmd_buffer, enc_info);
+      vid->enc_need_begin = false;
+   } else {
+      // temporal layers init
+      unsigned i = 0;
+      do {
+         if (vid->enc_need_rate_control) {
+            radv_enc_layer_select(cmd_buffer, i);
+            radv_enc_rc_layer_init(cmd_buffer, &vid->rc_layer_init[i]);
+            vid->enc_need_rate_control = false;
+         }
          radv_enc_layer_select(cmd_buffer, i);
-         radv_enc_rc_layer_init(cmd_buffer, &vid->rc_layer_init[i]);
-         vid->enc_need_rate_control = false;
-      }
-      radv_enc_layer_select(cmd_buffer, i);
-      radv_enc_rc_per_pic(cmd_buffer, enc_info, &vid->rc_per_pic[i]);
-   } while (++i < vid->rc_layer_control.num_temporal_layers);
+         radv_enc_rc_per_pic(cmd_buffer, enc_info, &vid->rc_per_pic[i]);
+      } while (++i < vid->rc_layer_control.num_temporal_layers);
+   }
 
    // encode headers
    // ctx
@@ -1724,7 +1740,7 @@ radv_vcn_encode_video(struct radv_cmd_buffer *cmd_buffer, const VkVideoEncodeInf
    radv_enc_bitstream(cmd_buffer, dst_buffer, enc_info->dstBufferOffset);
 
    // feedback
-   radv_enc_feedback(cmd_buffer);
+   radv_enc_feedback(cmd_buffer, feedback_query_va);
 
    // v2 encode statistics
    if (pdev->enc_hw_ver >= RADV_VIDEO_ENC_HW_2) {
@@ -1745,7 +1761,8 @@ radv_vcn_encode_video(struct radv_cmd_buffer *cmd_buffer, const VkVideoEncodeInf
 
    radeon_emit_direct(cmd_buffer->cs, enc->task_size_offset, enc->total_task_size);
 
-   destroy(cmd_buffer);
+   if (pdev->enc_hw_ver >= RADV_VIDEO_ENC_HW_4)
+      radv_vcn_sq_tail(cmd_buffer->cs, &cmd_buffer->video.sq);
 }
 
 static void
@@ -1922,6 +1939,9 @@ radv_video_patch_encode_session_parameters(struct vk_video_session_parameters *p
 {
    switch (params->op) {
    case VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR:
+      for (unsigned i = 0; i < params->h264_enc.h264_pps_count; i++) {
+         params->h264_enc.h264_pps[i].base.pic_init_qp_minus26 = 0;
+      }
       break;
    case VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR: {
       /*
@@ -1932,6 +1952,7 @@ radv_video_patch_encode_session_parameters(struct vk_video_session_parameters *p
       for (unsigned i = 0; i < params->h265_enc.h265_pps_count; i++) {
          params->h265_enc.h265_pps[i].base.flags.cu_qp_delta_enabled_flag = 1;
          params->h265_enc.h265_pps[i].base.diff_cu_qp_delta_depth = 0;
+         params->h265_enc.h265_pps[i].base.init_qp_minus26 = 0;
       }
       break;
    }
@@ -1971,6 +1992,13 @@ radv_GetEncodedVideoSessionParametersKHR(VkDevice device,
          char *data_ptr = pData ? (char *)pData + sps_size : NULL;
          vk_video_encode_h264_pps(pps, templ->vk.h264_enc.profile_idc == STD_VIDEO_H264_PROFILE_IDC_HIGH, size_limit,
                                   &pps_size, data_ptr);
+         if (pFeedbackInfo) {
+            struct VkVideoEncodeH264SessionParametersFeedbackInfoKHR *h264_feedback_info =
+               vk_find_struct(pFeedbackInfo->pNext, VIDEO_ENCODE_H264_SESSION_PARAMETERS_FEEDBACK_INFO_KHR);
+            pFeedbackInfo->hasOverrides = VK_TRUE;
+            if (h264_feedback_info)
+               h264_feedback_info->hasStdPPSOverrides = VK_TRUE;
+         }
       }
       total_size = sps_size + pps_size;
       break;
@@ -2017,26 +2045,6 @@ radv_GetEncodedVideoSessionParametersKHR(VkDevice device,
    return VK_SUCCESS;
 }
 
-void
-radv_video_enc_begin_coding(struct radv_cmd_buffer *cmd_buffer)
-{
-   struct radv_device *device = radv_cmd_buffer_device(cmd_buffer);
-   const struct radv_physical_device *pdev = radv_device_physical(device);
-   radeon_check_space(device->ws, cmd_buffer->cs, 1024);
-
-   if (pdev->enc_hw_ver >= RADV_VIDEO_ENC_HW_4)
-      radv_vcn_sq_header(cmd_buffer->cs, &cmd_buffer->video.sq, true);
-}
-
-void
-radv_video_enc_end_coding(struct radv_cmd_buffer *cmd_buffer)
-{
-   struct radv_device *device = radv_cmd_buffer_device(cmd_buffer);
-   const struct radv_physical_device *pdev = radv_device_physical(device);
-   if (pdev->enc_hw_ver >= RADV_VIDEO_ENC_HW_4)
-      radv_vcn_sq_tail(cmd_buffer->cs, &cmd_buffer->video.sq);
-}
-
 #define VCN_ENC_SESSION_SIZE 128 * 1024
 
 VkResult
