@@ -813,6 +813,7 @@ static const _mesa_glsl_extension _mesa_glsl_supported_extensions[] = {
    EXT(EXT_draw_buffers),
    EXT(EXT_draw_instanced),
    EXT(EXT_clip_cull_distance),
+   EXT(EXT_conservative_depth),
    EXT(EXT_geometry_point_size),
    EXT_AEP(EXT_geometry_shader),
    EXT(EXT_gpu_shader4),
@@ -1031,6 +1032,9 @@ _mesa_glsl_process_extension(const char *name, YYLTYPE *name_locp,
       }
    }
 
+   if (state->OVR_multiview2_enable)
+      state->OVR_multiview_enable = true;
+
    return true;
 }
 
@@ -2092,6 +2096,7 @@ set_shader_inout_layout(struct gl_shader *shader,
       break;
    }
 
+   shader->view_mask = state->view_mask;
    shader->bindless_sampler = state->bindless_sampler_specified;
    shader->bindless_image = state->bindless_image_specified;
    shader->bound_sampler = state->bound_sampler_specified;
@@ -2100,49 +2105,6 @@ set_shader_inout_layout(struct gl_shader *shader,
    shader->layer_viewport_relative = state->layer_viewport_relative;
 }
 
-/* src can be NULL if only the symbols found in the exec_list should be
- * copied
- */
-void
-_mesa_glsl_copy_symbols_from_table(struct exec_list *shader_ir,
-                                   struct glsl_symbol_table *src,
-                                   struct glsl_symbol_table *dest)
-{
-   foreach_in_list (ir_instruction, ir, shader_ir) {
-      switch (ir->ir_type) {
-      case ir_type_function:
-         dest->add_function((ir_function *) ir);
-         break;
-      case ir_type_variable: {
-         ir_variable *const var = (ir_variable *) ir;
-
-         if (var->data.mode != ir_var_temporary)
-            dest->add_variable(var);
-         break;
-      }
-      default:
-         break;
-      }
-   }
-
-   if (src != NULL) {
-      /* Explicitly copy the gl_PerVertex interface definitions because these
-       * are needed to check they are the same during the interstage link.
-       * They can’t necessarily be found via the exec_list because the members
-       * might not be referenced. The GL spec still requires that they match
-       * in that case.
-       */
-      const glsl_type *iface =
-         src->get_interface("gl_PerVertex", ir_var_shader_in);
-      if (iface)
-         dest->add_interface(glsl_get_type_name(iface), iface, ir_var_shader_in);
-
-      iface = src->get_interface("gl_PerVertex", ir_var_shader_out);
-      if (iface)
-         dest->add_interface(glsl_get_type_name(iface), iface, ir_var_shader_out);
-   }
-}
-
 extern "C" {
 
 static void
@@ -2215,10 +2177,9 @@ do_late_parsing_checks(struct _mesa_glsl_parse_state *state)
 }
 
 static void
-opt_shader_and_create_symbol_table(const struct gl_constants *consts,
-                                   const struct gl_extensions *exts,
-                                   struct glsl_symbol_table *source_symbols,
-                                   struct gl_shader *shader)
+opt_shader(const struct gl_constants *consts,
+           const struct gl_extensions *exts,
+           struct gl_shader *shader)
 {
    assert(shader->CompileStatus != COMPILE_FAILURE &&
           !shader->ir->is_empty());
@@ -2260,7 +2221,8 @@ opt_shader_and_create_symbol_table(const struct gl_constants *consts,
                           consts->GLSLHasHalfFloatPacking);
    do_mat_op_to_vec(shader->ir);
 
-   lower_instructions(shader->ir, exts->ARB_gpu_shader5);
+   lower_instructions(shader->ir, consts->ForceGLSLAbsSqrt,
+                      exts->ARB_gpu_shader5);
 
    do_vec_index_to_cond_assign(shader->ir);
 
@@ -2268,19 +2230,6 @@ opt_shader_and_create_symbol_table(const struct gl_constants *consts,
 
    /* Retain any live IR, but trash the rest. */
    reparent_ir(shader->ir, shader->ir);
-
-   /* Destroy the symbol table.  Create a new symbol table that contains only
-    * the variables and functions that still exist in the IR.  The symbol
-    * table will be used later during linking.
-    *
-    * There must NOT be any freed objects still referenced by the symbol
-    * table.  That could cause the linker to dereference freed memory.
-    *
-    * We don't have to worry about types or interface-types here because those
-    * are fly-weights that are looked up by glsl_type.
-    */
-   _mesa_glsl_copy_symbols_from_table(shader->ir, source_symbols,
-                                      shader->symbols);
 }
 
 static bool
@@ -2433,7 +2382,6 @@ _mesa_glsl_compile_shader(struct gl_context *ctx, struct gl_shader *shader,
    if (!state->error)
       set_shader_inout_layout(shader, state);
 
-   shader->symbols = new(shader->ir) glsl_symbol_table;
    shader->CompileStatus = state->error ? COMPILE_FAILURE : COMPILE_SUCCESS;
    shader->InfoLog = state->info_log;
    shader->Version = state->language_version;
@@ -2453,8 +2401,7 @@ _mesa_glsl_compile_shader(struct gl_context *ctx, struct gl_shader *shader,
       lower_builtins(shader->ir);
       assign_subroutine_indexes(state);
       lower_subroutine(shader->ir, state);
-      opt_shader_and_create_symbol_table(&ctx->Const, &ctx->Extensions,
-                                         state->symbols, shader);
+      opt_shader(&ctx->Const, &ctx->Extensions, shader);
    }
 
    if (!force_recompile) {
@@ -2499,8 +2446,7 @@ _mesa_glsl_compile_shader(struct gl_context *ctx, struct gl_shader *shader,
    if (shader->CompileStatus == COMPILE_SUCCESS) {
       memcpy(shader->compiled_source_blake3, source_blake3, BLAKE3_OUT_LEN);
 
-      shader->nir = glsl_to_nir(&ctx->Const, &shader->ir, NULL, shader->Stage,
-                                options->NirOptions, source_blake3);
+      shader->nir = glsl_to_nir(shader, options->NirOptions, source_blake3);
    }
 
    if (ctx->Cache && shader->CompileStatus == COMPILE_SUCCESS) {
@@ -2563,13 +2509,11 @@ do_common_optimization(exec_list *ir, bool linked,
       OPT(opt_flip_matrices, ir);
 
    OPT(do_dead_code_unlinked, ir);
-   OPT(do_dead_code_local, ir);
    OPT(do_tree_grafting, ir);
    OPT(do_minmax_prune, ir);
    OPT(do_rebalance_tree, ir);
    OPT(do_algebraic, ir, native_integers, options);
-   OPT(do_lower_jumps, ir, true, true, options->EmitNoMainReturn,
-       options->EmitNoCont);
+   OPT(do_lower_jumps, ir, true, options->EmitNoCont);
 
    /* If an optimization pass fails to preserve the invariant flag, calling
     * the pass only once earlier may result in incorrect code generation. Always call
