@@ -43,6 +43,7 @@
 /* clang-format on */
 #ifdef HAVE_LIBDRM
 #include <xf86drm.h>
+#include "platform_x11_dri3.h"
 #endif
 #include "util/bitscan.h"
 #include "util/macros.h"
@@ -50,15 +51,13 @@
 #include "util/log.h"
 #include <sys/stat.h>
 #include <sys/types.h>
-
+#include "loader_x11.h"
 #include "kopper_interface.h"
 #include "loader.h"
 #include "platform_x11.h"
 #include "drm-uapi/drm_fourcc.h"
+#include "dri_util.h"
 
-#ifdef HAVE_DRI3
-#include "platform_x11_dri3.h"
-#endif
 
 static EGLBoolean
 dri2_x11_swap_interval(_EGLDisplay *disp, _EGLSurface *surf, EGLint interval);
@@ -112,7 +111,7 @@ swrastDestroyDrawable(struct dri2_egl_display *dri2_dpy,
 }
 
 static bool
-x11_get_drawable_info(__DRIdrawable *draw, int *x, int *y, int *w, int *h,
+x11_get_drawable_info(struct dri_drawable *draw, int *x, int *y, int *w, int *h,
                       void *loaderPrivate)
 {
    struct dri2_egl_surface *dri2_surf = loaderPrivate;
@@ -145,7 +144,7 @@ x11_get_drawable_info(__DRIdrawable *draw, int *x, int *y, int *w, int *h,
 }
 
 static void
-swrastGetDrawableInfo(__DRIdrawable *draw, int *x, int *y, int *w, int *h,
+swrastGetDrawableInfo(struct dri_drawable *draw, int *x, int *y, int *w, int *h,
                       void *loaderPrivate)
 {
    *x = *y = *w = *h = 0;
@@ -153,7 +152,7 @@ swrastGetDrawableInfo(__DRIdrawable *draw, int *x, int *y, int *w, int *h,
 }
 
 static void
-swrastPutImage2(__DRIdrawable *draw, int op, int x, int y, int w, int h,
+swrastPutImage2(struct dri_drawable *draw, int op, int x, int y, int w, int h,
                 int stride, char *data, void *loaderPrivate)
 {
    struct dri2_egl_surface *dri2_surf = loaderPrivate;
@@ -218,7 +217,7 @@ swrastPutImage2(__DRIdrawable *draw, int op, int x, int y, int w, int h,
 }
 
 static void
-swrastPutImage(__DRIdrawable *draw, int op, int x, int y, int w, int h,
+swrastPutImage(struct dri_drawable *draw, int op, int x, int y, int w, int h,
                char *data, void *loaderPrivate)
 {
    struct dri2_egl_surface *dri2_surf = loaderPrivate;
@@ -227,7 +226,7 @@ swrastPutImage(__DRIdrawable *draw, int op, int x, int y, int w, int h,
 }
 
 static void
-swrastGetImage2(__DRIdrawable * read,
+swrastGetImage2(struct dri_drawable * read,
                 int x, int y, int w, int h, int stride,
                 char *data, void *loaderPrivate)
 {
@@ -267,7 +266,7 @@ swrastGetImage2(__DRIdrawable * read,
 }
 
 static void
-swrastGetImage(__DRIdrawable *read, int x, int y, int w, int h, char *data,
+swrastGetImage(struct dri_drawable *read, int x, int y, int w, int h, char *data,
                void *loaderPrivate)
 {
    struct dri2_egl_surface *dri2_surf = loaderPrivate;
@@ -276,7 +275,7 @@ swrastGetImage(__DRIdrawable *read, int x, int y, int w, int h, char *data,
 }
 
 static void
-swrastPutImageShm(__DRIdrawable * draw, int op,
+swrastPutImageShm(struct dri_drawable * draw, int op,
                   int x, int y, int w, int h, int stride,
                   int shmid, char *shmaddr, unsigned offset,
                   void *loaderPrivate)
@@ -327,7 +326,7 @@ swrastPutImageShm(__DRIdrawable * draw, int op,
 }
 
 static void
-swrastGetImageShm(__DRIdrawable * read,
+swrastGetImageShm(struct dri_drawable * read,
                   int x, int y, int w, int h,
                   int shmid, void *loaderPrivate)
 {
@@ -419,7 +418,7 @@ dri2_x11_create_surface(_EGLDisplay *disp, EGLint type, _EGLConfig *conf,
    xcb_get_geometry_cookie_t cookie;
    xcb_get_geometry_reply_t *reply;
    xcb_generic_error_t *error;
-   const __DRIconfig *config;
+   const struct dri_config *config;
 
    dri2_surf = calloc(1, sizeof *dri2_surf);
    if (!dri2_surf) {
@@ -477,7 +476,8 @@ dri2_x11_create_surface(_EGLDisplay *disp, EGLint type, _EGLConfig *conf,
    if (!dri2_create_drawable(dri2_dpy, config, dri2_surf, dri2_surf))
       goto cleanup_pixmap;
 
-   if (dri2_dpy->dri2) {
+#ifdef HAVE_X11_DRI2
+   if (!dri2_dpy->swrast) {
       xcb_void_cookie_t cookie;
       int conn_error;
 
@@ -498,7 +498,9 @@ dri2_x11_create_surface(_EGLDisplay *disp, EGLint type, _EGLConfig *conf,
          free(error);
          goto cleanup_dri_drawable;
       }
-   } else {
+   } else
+#endif
+   {
       if (type == EGL_PBUFFER_BIT) {
          dri2_surf->depth = conf->BufferSize;
       }
@@ -511,7 +513,7 @@ dri2_x11_create_surface(_EGLDisplay *disp, EGLint type, _EGLConfig *conf,
    return &dri2_surf->base;
 
 cleanup_dri_drawable:
-   dri2_dpy->core->destroyDrawable(dri2_surf->dri_drawable);
+   driDestroyDrawable(dri2_surf->dri_drawable);
 cleanup_pixmap:
    if (type == EGL_PBUFFER_BIT)
       xcb_free_pixmap(dri2_dpy->conn, dri2_surf->drawable);
@@ -568,12 +570,13 @@ dri2_x11_destroy_surface(_EGLDisplay *disp, _EGLSurface *surf)
    struct dri2_egl_display *dri2_dpy = dri2_egl_display(disp);
    struct dri2_egl_surface *dri2_surf = dri2_egl_surface(surf);
 
-   dri2_dpy->core->destroyDrawable(dri2_surf->dri_drawable);
+   driDestroyDrawable(dri2_surf->dri_drawable);
 
-   if (dri2_dpy->dri2) {
+   if (!dri2_dpy->swrast) {
+#ifdef HAVE_X11_DRI2
       xcb_dri2_destroy_drawable(dri2_dpy->conn, dri2_surf->drawable);
+#endif
    } else {
-      assert(dri2_dpy->swrast);
       swrastDestroyDrawable(dri2_dpy, dri2_surf);
    }
 
@@ -602,7 +605,7 @@ dri2_query_surface(_EGLDisplay *disp, _EGLSurface *surf, EGLint attribute,
    struct dri2_egl_surface *dri2_surf = dri2_egl_surface(surf);
    int x, y, w, h;
 
-   __DRIdrawable *drawable = dri2_dpy->vtbl->get_dri_drawable(surf);
+   struct dri_drawable *drawable = dri2_dpy->vtbl->get_dri_drawable(surf);
 
    switch (attribute) {
    case EGL_WIDTH:
@@ -611,8 +614,8 @@ dri2_query_surface(_EGLDisplay *disp, _EGLSurface *surf, EGLint attribute,
          bool changed = surf->Width != w || surf->Height != h;
          surf->Width = w;
          surf->Height = h;
-         if (changed && dri2_dpy->flush)
-            dri2_dpy->flush->invalidate(drawable);
+         if (changed && !dri2_dpy->swrast_not_kms)
+            dri_invalidate_drawable(drawable);
       }
       break;
    default:
@@ -621,6 +624,7 @@ dri2_query_surface(_EGLDisplay *disp, _EGLSurface *surf, EGLint attribute,
    return _eglQuerySurface(disp, surf, attribute, value);
 }
 
+#ifdef HAVE_X11_DRI2
 /**
  * Process list of buffer received from the server
  *
@@ -667,7 +671,7 @@ dri2_x11_process_buffers(struct dri2_egl_surface *dri2_surf,
 }
 
 static __DRIbuffer *
-dri2_x11_get_buffers(__DRIdrawable *driDrawable, int *width, int *height,
+dri2_x11_get_buffers(struct dri_drawable *driDrawable, int *width, int *height,
                      unsigned int *attachments, int count, int *out_count,
                      void *loaderPrivate)
 {
@@ -702,7 +706,7 @@ dri2_x11_get_buffers(__DRIdrawable *driDrawable, int *width, int *height,
 }
 
 static __DRIbuffer *
-dri2_x11_get_buffers_with_format(__DRIdrawable *driDrawable, int *width,
+dri2_x11_get_buffers_with_format(struct dri_drawable *driDrawable, int *width,
                                  int *height, unsigned int *attachments,
                                  int count, int *out_count, void *loaderPrivate)
 {
@@ -736,7 +740,7 @@ dri2_x11_get_buffers_with_format(__DRIdrawable *driDrawable, int *width,
 }
 
 static void
-dri2_x11_flush_front_buffer(__DRIdrawable *driDrawable, void *loaderPrivate)
+dri2_x11_flush_front_buffer(struct dri_drawable *driDrawable, void *loaderPrivate)
 {
    (void)driDrawable;
 
@@ -911,6 +915,7 @@ dri2_x11_authenticate(_EGLDisplay *disp, uint32_t id)
 
    return dri2_x11_do_authenticate(dri2_dpy, id);
 }
+#endif
 
 static void
 dri2_x11_add_configs_for_visuals(struct dri2_egl_display *dri2_dpy,
@@ -971,11 +976,11 @@ dri2_x11_add_configs_for_visuals(struct dri2_egl_display *dri2_dpy,
          };
 
          for (int j = 0; dri2_dpy->driver_configs[j]; j++) {
-            const __DRIconfig *config = dri2_dpy->driver_configs[j];
+            const struct dri_config *config = dri2_dpy->driver_configs[j];
             int shifts[4];
             unsigned int sizes[4];
 
-            dri2_get_shifts_and_sizes(dri2_dpy->core, config, shifts, sizes);
+            dri2_get_shifts_and_sizes(config, shifts, sizes);
 
             if (memcmp(shifts, rgb_shifts, sizeof(rgb_shifts)) != 0 ||
                 memcmp(sizes, rgb_sizes, sizeof(rgb_sizes)) != 0) {
@@ -1012,6 +1017,7 @@ dri2_x11_add_configs_for_visuals(struct dri2_egl_display *dri2_dpy,
    }
 }
 
+#ifdef HAVE_X11_DRI2
 static EGLBoolean
 dri2_copy_region(_EGLDisplay *disp, _EGLSurface *draw,
                  xcb_xfixes_region_t region)
@@ -1026,7 +1032,7 @@ dri2_copy_region(_EGLDisplay *disp, _EGLSurface *draw,
       return EGL_TRUE;
 
    assert(!dri2_dpy->kopper);
-   dri2_dpy->flush->flush(dri2_surf->dri_drawable);
+   dri_flush_drawable(dri2_surf->dri_drawable);
 
    if (dri2_surf->have_fake_front)
       render_attachment = XCB_DRI2_ATTACHMENT_BUFFER_FAKE_FRONT_LEFT;
@@ -1085,11 +1091,11 @@ dri2_x11_swap_buffers_msc(_EGLDisplay *disp, _EGLSurface *draw, int64_t msc,
     * happened.  The driver should still be using the viewport hack to catch
     * window resizes.
     */
-   if (dri2_dpy->flush->base.version >= 3 && dri2_dpy->flush->invalidate)
-      dri2_dpy->flush->invalidate(dri2_surf->dri_drawable);
+   dri_invalidate_drawable(dri2_surf->dri_drawable);
 
    return swap_count;
 }
+#endif
 
 static EGLBoolean
 dri2_x11_swap_buffers(_EGLDisplay *disp, _EGLSurface *draw)
@@ -1103,22 +1109,25 @@ dri2_x11_swap_buffers(_EGLDisplay *disp, _EGLSurface *draw)
        *     "The contents of ancillary buffers are always undefined
        *      after calling eglSwapBuffers."
        */
-      dri2_dpy->kopper->swapBuffers(dri2_surf->dri_drawable,
+      kopperSwapBuffers(dri2_surf->dri_drawable,
                                     __DRI2_FLUSH_INVALIDATE_ANCILLARY);
       return EGL_TRUE;
-   } else if (!dri2_dpy->flush) {
+   } else if (dri2_dpy->swrast) {
       /* aka the swrast path, which does the swap in the gallium driver. */
-      dri2_dpy->core->swapBuffers(dri2_surf->dri_drawable);
+      driSwapBuffers(dri2_surf->dri_drawable);
       return EGL_TRUE;
    }
 
+#ifdef HAVE_X11_DRI2
    if (dri2_x11_swap_buffers_msc(disp, draw, 0, 0, 0) == -1) {
       /* Swap failed with a window drawable. */
       return _eglError(EGL_BAD_NATIVE_WINDOW, __func__);
    }
+#endif
    return EGL_TRUE;
 }
 
+#ifdef HAVE_X11_DRI2
 static EGLBoolean
 dri2_x11_swap_buffers_region(_EGLDisplay *disp, _EGLSurface *draw,
                              EGLint numRects, const EGLint *rects)
@@ -1159,6 +1168,10 @@ dri2_x11_post_sub_buffer(_EGLDisplay *disp, _EGLSurface *draw, EGLint x,
 
    return dri2_x11_swap_buffers_region(disp, draw, 1, rect);
 }
+#else
+#define dri2_x11_swap_buffers_region NULL
+#define dri2_x11_post_sub_buffer NULL
+#endif
 
 static EGLBoolean
 dri2_x11_kopper_swap_buffers_with_damage(_EGLDisplay *disp, _EGLSurface *draw,
@@ -1170,14 +1183,14 @@ dri2_x11_kopper_swap_buffers_with_damage(_EGLDisplay *disp, _EGLSurface *draw,
    assert(dri2_dpy->kopper);
    if (numRects) {
       if (dri2_dpy->kopper)
-         dri2_dpy->kopper->swapBuffersWithDamage(dri2_surf->dri_drawable, __DRI2_FLUSH_INVALIDATE_ANCILLARY, numRects, rects);
+         kopperSwapBuffersWithDamage(dri2_surf->dri_drawable, __DRI2_FLUSH_INVALIDATE_ANCILLARY, numRects, rects);
       else
-         dri2_dpy->core->swapBuffersWithDamage(dri2_surf->dri_drawable, numRects, rects);
+         driSwapBuffersWithDamage(dri2_surf->dri_drawable, numRects, rects);
    } else {
       if (dri2_dpy->kopper)
-         dri2_dpy->kopper->swapBuffers(dri2_surf->dri_drawable, __DRI2_FLUSH_INVALIDATE_ANCILLARY);
+         kopperSwapBuffers(dri2_surf->dri_drawable, __DRI2_FLUSH_INVALIDATE_ANCILLARY);
       else
-         dri2_dpy->core->swapBuffers(dri2_surf->dri_drawable);
+         driSwapBuffers(dri2_surf->dri_drawable);
    }
    return EGL_TRUE;
 }
@@ -1186,12 +1199,11 @@ static EGLBoolean
 dri2_x11_swap_buffers_with_damage(_EGLDisplay *disp, _EGLSurface *draw,
                                   const EGLint *rects, EGLint numRects)
 {
-   struct dri2_egl_display *dri2_dpy = dri2_egl_display(disp);
    struct dri2_egl_surface *dri2_surf = dri2_egl_surface(draw);
    if (numRects)
-      dri2_dpy->core->swapBuffersWithDamage(dri2_surf->dri_drawable, numRects, rects);
+      driSwapBuffersWithDamage(dri2_surf->dri_drawable, numRects, rects);
    else
-      dri2_dpy->core->swapBuffers(dri2_surf->dri_drawable);
+      driSwapBuffers(dri2_surf->dri_drawable);
    return EGL_TRUE;
 }
 
@@ -1202,12 +1214,14 @@ dri2_x11_swap_interval(_EGLDisplay *disp, _EGLSurface *surf, EGLint interval)
    struct dri2_egl_surface *dri2_surf = dri2_egl_surface(surf);
 
    if (dri2_dpy->kopper) {
-      dri2_dpy->kopper->setSwapInterval(dri2_surf->dri_drawable, interval);
+      kopperSetSwapInterval(dri2_surf->dri_drawable, interval);
       return EGL_TRUE;
    }
 
+#ifdef HAVE_X11_DRI2
    if (dri2_dpy->swap_available)
       xcb_dri2_swap_interval(dri2_dpy->conn, dri2_surf->drawable, interval);
+#endif
 
    return EGL_TRUE;
 }
@@ -1224,8 +1238,8 @@ dri2_x11_copy_buffers(_EGLDisplay *disp, _EGLSurface *surf,
    STATIC_ASSERT(sizeof(uintptr_t) == sizeof(native_pixmap_target));
    target = (uintptr_t)native_pixmap_target;
 
-   if (dri2_dpy->flush)
-      dri2_dpy->flush->flush(dri2_surf->dri_drawable);
+   if (!dri2_dpy->swrast_not_kms)
+      dri_flush_drawable(dri2_surf->dri_drawable);
    else {
       /* This should not be a swapBuffers, because it could present an
        * incomplete frame, and it could invalidate the back buffer if it's not
@@ -1233,7 +1247,7 @@ dri2_x11_copy_buffers(_EGLDisplay *disp, _EGLSurface *surf,
        * okay-ish on swrast because those aren't invalidating the back buffer on
        * swap.
        */
-      dri2_dpy->core->swapBuffers(dri2_surf->dri_drawable);
+      driSwapBuffers(dri2_surf->dri_drawable);
    }
 
    gc = xcb_generate_id(dri2_dpy->conn);
@@ -1266,6 +1280,7 @@ dri2_fourcc_for_depth(struct dri2_egl_display *dri2_dpy, uint32_t depth)
    }
 }
 
+#ifdef HAVE_X11_DRI2
 static _EGLImage *
 dri2_create_image_khr_pixmap(_EGLDisplay *disp, _EGLContext *ctx,
                              EGLClientBuffer buffer, const EGLint *attr_list)
@@ -1330,10 +1345,11 @@ dri2_create_image_khr_pixmap(_EGLDisplay *disp, _EGLContext *ctx,
 
    _eglInitImage(&dri2_img->base, disp);
 
-   dri2_img->dri_image = dri2_dpy->image->createImageFromNames(
+   int offset = 0;
+   dri2_img->dri_image = dri2_from_names(
       dri2_dpy->dri_screen_render_gpu, buffers_reply->width,
       buffers_reply->height, fourcc, (int *) &buffers[0].name, 1,
-      (int *) &buffers[0].pitch, 0, dri2_img);
+      (int *) &buffers[0].pitch, &offset, dri2_img);
 
    free(buffers_reply);
    free(geometry_reply);
@@ -1376,6 +1392,7 @@ dri2_x11_get_sync_values(_EGLDisplay *display, _EGLSurface *surface,
 
    return EGL_TRUE;
 }
+#endif
 
 static int
 box_intersection_area(int16_t a_x, int16_t a_y, int16_t a_width,
@@ -1394,7 +1411,7 @@ dri2_x11_get_msc_rate(_EGLDisplay *display, _EGLSurface *surface,
 {
    struct dri2_egl_display *dri2_dpy = dri2_egl_display(display);
 
-#ifdef HAVE_DRI3
+#ifdef HAVE_LIBDRM
    loader_update_screen_resources(&dri2_dpy->screen_resources);
 
    if (dri2_dpy->screen_resources.num_crtcs == 0) {
@@ -1416,6 +1433,7 @@ dri2_x11_get_msc_rate(_EGLDisplay *display, _EGLSurface *surface,
    *denominator = 1;
 #endif
 
+
    /* In a multi-monitor setup, look at each CRTC and perform a box
     * intersection between the CRTC and surface.  Use the CRTC whose
     * box intersection has the largest area.
@@ -1437,7 +1455,7 @@ dri2_x11_get_msc_rate(_EGLDisplay *display, _EGLSurface *surface,
       return EGL_FALSE;
    }
 
-#ifdef HAVE_DRI3
+#ifdef HAVE_LIBDRM
    int area = 0;
 
    for (unsigned c = 0; c < dri2_dpy->screen_resources.num_crtcs; c++) {
@@ -1453,7 +1471,6 @@ dri2_x11_get_msc_rate(_EGLDisplay *display, _EGLSurface *surface,
       }
    }
 #endif
-
    /* If the window is entirely off-screen, then area will still be 0.
     * We defaulted to the first CRTC in the list's refresh rate, earlier.
     */
@@ -1469,7 +1486,7 @@ dri2_kopper_swap_interval(_EGLDisplay *disp, _EGLSurface *surf, EGLint interval)
 
    /* This can legitimately be null for lavapipe */
    if (dri2_dpy->kopper)
-      dri2_dpy->kopper->setSwapInterval(dri2_surf->dri_drawable, interval);
+      kopperSetSwapInterval(dri2_surf->dri_drawable, interval);
 
    return EGL_TRUE;
 }
@@ -1505,9 +1522,9 @@ dri2_kopper_query_buffer_age(_EGLDisplay *disp, _EGLSurface *surf)
 
    /* This can legitimately be null for lavapipe */
    if (dri2_dpy->kopper)
-      return dri2_dpy->kopper->queryBufferAge(dri2_surf->dri_drawable);
+      return kopperQueryBufferAge(dri2_surf->dri_drawable);
    else
-      return dri2_dpy->swrast->queryBufferAge(dri2_surf->dri_drawable);
+      return driSWRastQueryBufferAge(dri2_surf->dri_drawable);
 
    return 0;
 }
@@ -1515,11 +1532,9 @@ dri2_kopper_query_buffer_age(_EGLDisplay *disp, _EGLSurface *surf)
 static EGLint
 dri2_swrast_query_buffer_age(_EGLDisplay *disp, _EGLSurface *surf)
 {
-   struct dri2_egl_display *dri2_dpy = dri2_egl_display(disp);
    struct dri2_egl_surface *dri2_surf = dri2_egl_surface(surf);
 
-   assert(dri2_dpy->swrast);
-   return dri2_dpy->swrast->queryBufferAge(dri2_surf->dri_drawable);
+   return driSWRastQueryBufferAge(dri2_surf->dri_drawable);
 }
 
 static const struct dri2_egl_display_vtbl dri2_x11_swrast_display_vtbl = {
@@ -1561,6 +1576,7 @@ static const struct dri2_egl_display_vtbl dri2_x11_kopper_display_vtbl = {
    .get_dri_drawable = dri2_surface_get_dri_drawable,
 };
 
+#ifdef HAVE_X11_DRI2
 static const struct dri2_egl_display_vtbl dri2_x11_display_vtbl = {
    .authenticate = dri2_x11_authenticate,
    .create_window_surface = dri2_x11_create_window_surface,
@@ -1578,6 +1594,7 @@ static const struct dri2_egl_display_vtbl dri2_x11_display_vtbl = {
    .get_msc_rate = dri2_x11_get_msc_rate,
    .get_dri_drawable = dri2_surface_get_dri_drawable,
 };
+#endif
 
 static const __DRIswrastLoaderExtension swrast_loader_extension = {
    .base = {__DRI_SWRAST_LOADER, 1},
@@ -1629,6 +1646,14 @@ static const __DRIkopperLoaderExtension kopper_loader_extension = {
    .SetSurfaceCreateInfo = kopperSetSurfaceCreateInfo,
 };
 
+static const __DRIextension *kopper_loader_extensions[] = {
+   &swrast_loader_extension.base,
+   &image_lookup_extension.base,
+   &kopper_loader_extension.base,
+   &use_invalidate.base,
+   NULL,
+};
+
 static const __DRIextension *swrast_loader_extensions[] = {
    &swrast_loader_extension.base,
    &image_lookup_extension.base,
@@ -1758,6 +1783,23 @@ check_xshm(struct dri2_egl_display *dri2_dpy)
    return ret;
 }
 
+static EGLBoolean
+dri2_x11_check_multibuffers(_EGLDisplay *disp)
+{
+   struct dri2_egl_display *dri2_dpy = dri2_egl_display(disp);
+
+#ifdef HAVE_X11_DRM
+   bool err;
+   dri2_dpy->multibuffers_available = x11_dri3_check_multibuffer(dri2_dpy->conn, &err, &dri2_dpy->explicit_modifiers);
+
+   if (disp->Options.Zink && !disp->Options.ForceSoftware &&
+       !dri2_dpy->multibuffers_available &&
+       !dri2_dpy->kopper_without_modifiers)
+      return EGL_FALSE;
+#endif
+
+   return EGL_TRUE;
+}
 
 static EGLBoolean
 dri2_initialize_x11_swrast(_EGLDisplay *disp)
@@ -1774,28 +1816,32 @@ dri2_initialize_x11_swrast(_EGLDisplay *disp)
     * here will allow is to simply free the memory at dri2_terminate().
     */
    dri2_dpy->driver_name = strdup(disp->Options.Zink ? "zink" : "swrast");
-#ifdef HAVE_DRI3
+
+#ifdef HAVE_LIBDRM
    if (disp->Options.Zink &&
        !debug_get_bool_option("LIBGL_DRI3_DISABLE", false) &&
-       !debug_get_bool_option("LIBGL_KOPPER_DRI2", false))
+       (!disp->Options.Zink || !debug_get_bool_option("LIBGL_KOPPER_DRI2", false)))
       dri3_x11_connect(dri2_dpy, disp->Options.Zink, disp->Options.ForceSoftware);
 #endif
-   if (!dri2_load_driver_swrast(disp))
+
+   if (!dri2_load_driver(disp))
       goto cleanup;
 
-   if (check_xshm(dri2_dpy)) {
+   if (disp->Options.Zink && !disp->Options.ForceSoftware) {
+      dri2_dpy->loader_extensions = kopper_loader_extensions;
+   } else if (check_xshm(dri2_dpy)) {
       dri2_dpy->loader_extensions = swrast_loader_shm_extensions;
    } else {
       dri2_dpy->loader_extensions = swrast_loader_extensions;
    }
 
+   if (!dri2_x11_check_multibuffers(disp))
+      goto cleanup;
+
    if (!dri2_create_screen(disp))
       goto cleanup;
 
-   if (!dri2_setup_extensions(disp))
-      goto cleanup;
-
-   if (!dri2_setup_device(disp, true)) {
+   if (!dri2_setup_device(disp, disp->Options.ForceSoftware || dri2_dpy->kopper_without_modifiers)) {
       _eglError(EGL_NOT_INITIALIZED, "DRI2: failed to setup EGLDevice");
       goto cleanup;
    }
@@ -1815,7 +1861,7 @@ dri2_initialize_x11_swrast(_EGLDisplay *disp)
       disp->Extensions.CHROMIUM_sync_control = EGL_TRUE;
       disp->Extensions.EXT_swap_buffers_with_damage = !!dri2_dpy->kopper;
 
-#ifdef HAVE_DRI3
+#ifdef HAVE_LIBDRM
       if (dri2_dpy->multibuffers_available)
          dri2_set_WL_bind_wayland_display(disp);
 #endif
@@ -1842,8 +1888,7 @@ cleanup:
    return EGL_FALSE;
 }
 
-#ifdef HAVE_DRI3
-
+#ifdef HAVE_LIBDRM
 static const __DRIextension *dri3_image_loader_extensions[] = {
    &dri3_image_loader_extension.base,
    &image_lookup_extension.base,
@@ -1867,7 +1912,7 @@ dri2_initialize_x11_dri3(_EGLDisplay *disp)
    if (status != DRI2_EGL_DRIVER_LOADED)
       goto cleanup;
 
-   if (!dri2_load_driver_dri3(disp))
+   if (!dri2_load_driver(disp))
       goto cleanup;
 
    dri2_dpy->loader_extensions = dri3_image_loader_extensions;
@@ -1875,10 +1920,10 @@ dri2_initialize_x11_dri3(_EGLDisplay *disp)
    dri2_dpy->swap_available = true;
    dri2_dpy->invalidate_available = true;
 
-   if (!dri2_create_screen(disp))
+   if (!dri2_x11_check_multibuffers(disp))
       goto cleanup;
 
-   if (!dri2_setup_extensions(disp))
+   if (!dri2_create_screen(disp))
       goto cleanup;
 
    if (!dri2_setup_device(disp, false)) {
@@ -1905,13 +1950,6 @@ dri2_initialize_x11_dri3(_EGLDisplay *disp)
    loader_init_screen_resources(&dri2_dpy->screen_resources, dri2_dpy->conn,
                                 dri2_dpy->screen);
 
-   dri2_dpy->loader_dri3_ext.core = dri2_dpy->core;
-   dri2_dpy->loader_dri3_ext.image_driver = dri2_dpy->image_driver;
-   dri2_dpy->loader_dri3_ext.flush = dri2_dpy->flush;
-   dri2_dpy->loader_dri3_ext.tex_buffer = dri2_dpy->tex_buffer;
-   dri2_dpy->loader_dri3_ext.image = dri2_dpy->image;
-   dri2_dpy->loader_dri3_ext.config = dri2_dpy->config;
-
    /* Fill vtbl last to prevent accidentally calling virtual function during
     * initialization.
     */
@@ -1929,6 +1967,7 @@ cleanup:
 }
 #endif
 
+#ifdef HAVE_X11_DRI2
 static const __DRIdri2LoaderExtension dri2_loader_extension_old = {
    .base = {__DRI_DRI2_LOADER, 2},
 
@@ -1984,10 +2023,10 @@ dri2_initialize_x11_dri2(_EGLDisplay *disp)
    dri2_dpy->swap_available = (dri2_dpy->dri2_minor >= 2);
    dri2_dpy->invalidate_available = (dri2_dpy->dri2_minor >= 3);
 
-   if (!dri2_create_screen(disp))
+   if (!dri2_x11_check_multibuffers(disp))
       goto cleanup;
 
-   if (!dri2_setup_extensions(disp))
+   if (!dri2_create_screen(disp))
       goto cleanup;
 
    if (!dri2_setup_device(disp, false)) {
@@ -2023,6 +2062,7 @@ cleanup:
    dri2_display_destroy(disp);
    return EGL_FALSE;
 }
+#endif
 
 EGLBoolean
 dri2_initialize_x11(_EGLDisplay *disp)
@@ -2032,7 +2072,7 @@ dri2_initialize_x11(_EGLDisplay *disp)
        (disp->Options.Zink && !debug_get_bool_option("LIBGL_KOPPER_DISABLE", false)))
       return dri2_initialize_x11_swrast(disp);
 
-#ifdef HAVE_DRI3
+#ifdef HAVE_LIBDRM
    if (!debug_get_bool_option("LIBGL_DRI3_DISABLE", false)) {
       status = dri2_initialize_x11_dri3(disp);
       if (status == DRI2_EGL_DRIVER_LOADED)
@@ -2040,10 +2080,12 @@ dri2_initialize_x11(_EGLDisplay *disp)
    }
 #endif
 
+#ifdef HAVE_X11_DRI2
    if (!debug_get_bool_option("LIBGL_DRI2_DISABLE", false) &&
        status != DRI2_EGL_DRIVER_PREFER_ZINK)
       if (dri2_initialize_x11_dri2(disp))
          return EGL_TRUE;
+#endif
 
    return EGL_FALSE;
 }
@@ -2051,7 +2093,7 @@ dri2_initialize_x11(_EGLDisplay *disp)
 void
 dri2_teardown_x11(struct dri2_egl_display *dri2_dpy)
 {
-#ifdef HAVE_DRI3
+#ifdef HAVE_LIBDRM
    if (dri2_dpy->dri2_major >= 3)
       loader_destroy_screen_resources(&dri2_dpy->screen_resources);
 #endif
