@@ -50,6 +50,7 @@
 #include "compiler/nir/nir_serialize.h"
 #include "util/pan_lower_framebuffer.h"
 #include "decode.h"
+#include "pan_device.h"
 #include "pan_fence.h"
 #include "pan_screen.h"
 #include "pan_util.h"
@@ -68,6 +69,8 @@ panfrost_clear(struct pipe_context *pipe, unsigned buffers,
     */
    struct panfrost_context *ctx = pan_context(pipe);
    struct panfrost_batch *batch = panfrost_get_batch_for_fbo(ctx);
+   if (!batch)
+      return;
 
    /* At the start of the batch, we can clear for free */
    if (batch->draw_count == 0) {
@@ -159,7 +162,7 @@ panfrost_set_blend_color(struct pipe_context *pipe,
 
 /* Create a final blend given the context */
 
-mali_ptr
+uint64_t
 panfrost_get_blend(struct panfrost_batch *batch, unsigned rti,
                    struct panfrost_bo **bo, unsigned *shader_offset)
 {
@@ -173,6 +176,7 @@ panfrost_get_blend(struct panfrost_batch *batch, unsigned rti,
    /* Use fixed-function if the equation permits, the format is blendable,
     * and no more than one unique constant is accessed */
    if (info.fixed_function && dev->blendable_formats[fmt].internal &&
+       !blend->base.alpha_to_one &&
        pan_blend_is_homogenous_constant(info.constant_mask,
                                         ctx->blend_color.color)) {
       return 0;
@@ -189,7 +193,7 @@ panfrost_get_blend(struct panfrost_batch *batch, unsigned rti,
     * conversion descriptor in the internal blend descriptor. (Midgard
     * requires a blend shader even for this case.)
     */
-   if (dev->arch >= 6 && info.opaque)
+   if (dev->arch >= 6 && info.opaque && !blend->base.alpha_to_one)
       return 0;
 
    /* Otherwise, we need to grab a shader */
@@ -205,6 +209,10 @@ panfrost_get_blend(struct panfrost_batch *batch, unsigned rti,
    if (!(*bo)) {
       *bo = panfrost_batch_create_bo(batch, 4096, PAN_BO_EXECUTE,
                                      PIPE_SHADER_FRAGMENT, "Blend shader");
+      if (!(*bo)) {
+         mesa_loge("failed to allocate blend-shader");
+         return 0;
+      }
    }
 
    struct panfrost_compiled_shader *ss = ctx->prog[PIPE_SHADER_FRAGMENT];
@@ -268,15 +276,12 @@ panfrost_set_shader_images(struct pipe_context *pctx,
       return;
    }
 
-   /* Bind start_slot...start_slot+count */
+   /* fix up AFBC/AFRC; we do this before starting setup because it involves a
+      blit, so we need a consistent image state */
    for (int i = 0; i < count; i++) {
       const struct pipe_image_view *image = &iviews[i];
-      SET_BIT(ctx->image_mask[shader], 1 << (start_slot + i), image->resource);
-
-      if (!image->resource) {
-         util_copy_image_view(&ctx->images[shader][start_slot + i], NULL);
+      if (!image->resource)
          continue;
-      }
 
       struct panfrost_resource *rsrc = pan_resource(image->resource);
 
@@ -288,7 +293,16 @@ panfrost_set_shader_images(struct pipe_context *pctx,
             ctx, rsrc, DRM_FORMAT_MOD_ARM_16X16_BLOCK_U_INTERLEAVED, true,
             "Shader image");
       }
+   }
 
+   /* Bind start_slot...start_slot+count */
+   for (int i = 0; i < count; i++) {
+      const struct pipe_image_view *image = &iviews[i];
+      SET_BIT(ctx->image_mask[shader], 1 << (start_slot + i), image->resource);
+      if (!image->resource) {
+         util_copy_image_view(&ctx->images[shader][start_slot + i], NULL);
+         continue;
+      }
       util_copy_image_view(&ctx->images[shader][start_slot + i], image);
    }
 
@@ -584,11 +598,47 @@ panfrost_destroy(struct pipe_context *pipe)
 static struct pipe_query *
 panfrost_create_query(struct pipe_context *pipe, unsigned type, unsigned index)
 {
+   struct panfrost_context *ctx = pan_context(pipe);
+   struct panfrost_device *dev = pan_device(ctx->base.screen);
    struct panfrost_query *q = rzalloc(pipe, struct panfrost_query);
 
    q->type = type;
    q->index = index;
 
+   unsigned size = 0;
+
+   switch (q->type) {
+   case PIPE_QUERY_OCCLUSION_COUNTER:
+   case PIPE_QUERY_OCCLUSION_PREDICATE:
+   case PIPE_QUERY_OCCLUSION_PREDICATE_CONSERVATIVE:
+      size = sizeof(uint64_t) * dev->core_id_range;
+      break;
+   case PIPE_QUERY_TIMESTAMP:
+      size = sizeof(uint64_t);
+      break;
+   case PIPE_QUERY_TIME_ELAPSED:
+      size = sizeof(uint64_t) * 2;
+      break;
+   default:
+      break;
+   }
+
+   /* Allocate a resource for the query results to be stored */
+   if (size > 0) {
+      q->rsrc =
+         pipe_buffer_create(ctx->base.screen, PIPE_BIND_QUERY_BUFFER, 0, size);
+
+      if (!q->rsrc) {
+         ralloc_free(q);
+         return NULL;
+      }
+
+      /* Default to 0 if nothing at all drawn. */
+      uint8_t *zeroes = alloca(size);
+      memset(zeroes, 0, size);
+      pipe_buffer_write(pipe, q->rsrc, 0, size, zeroes);
+   }
+
    return (struct pipe_query *)q;
 }
 
@@ -603,6 +653,17 @@ panfrost_destroy_query(struct pipe_context *pipe, struct pipe_query *q)
    ralloc_free(q);
 }
 
+static void
+panfrost_emit_write_timestamp(struct panfrost_context *ctx,
+                              struct panfrost_resource *rsrc, unsigned offset,
+                              const char *reason)
+{
+   struct panfrost_screen *screen = pan_screen(ctx->base.screen);
+   struct panfrost_batch *batch = panfrost_get_fresh_batch_for_fbo(ctx, reason);
+
+   screen->vtbl.emit_write_timestamp(batch, rsrc, offset);
+}
+
 static bool
 panfrost_begin_query(struct pipe_context *pipe, struct pipe_query *q)
 {
@@ -616,12 +677,6 @@ panfrost_begin_query(struct pipe_context *pipe, struct pipe_query *q)
    case PIPE_QUERY_OCCLUSION_PREDICATE_CONSERVATIVE: {
       unsigned size = sizeof(uint64_t) * dev->core_id_range;
 
-      /* Allocate a resource for the query results to be stored */
-      if (!query->rsrc) {
-         query->rsrc = pipe_buffer_create(ctx->base.screen,
-                                          PIPE_BIND_QUERY_BUFFER, 0, size);
-      }
-
       /* Default to 0 if nothing at all drawn. */
       uint8_t *zeroes = alloca(size);
       memset(zeroes, 0, size);
@@ -633,6 +688,17 @@ panfrost_begin_query(struct pipe_context *pipe, struct pipe_query *q)
       break;
    }
 
+   case PIPE_QUERY_TIMESTAMP:
+   case PIPE_QUERY_TIMESTAMP_DISJOINT:
+      /* This is a no op on query start */
+      break;
+
+   case PIPE_QUERY_TIME_ELAPSED: {
+      struct panfrost_resource *rsrc = pan_resource(query->rsrc);
+      panfrost_emit_write_timestamp(ctx, rsrc, 0, "TIME_ELAPSED begin_query");
+      break;
+   }
+
       /* Geometry statistics are computed in the driver. XXX: geom/tess
        * shaders.. */
 
@@ -648,7 +714,6 @@ panfrost_begin_query(struct pipe_context *pipe, struct pipe_query *q)
       break;
 
    default:
-      /* TODO: timestamp queries, etc? */
       break;
    }
 
@@ -668,6 +733,20 @@ panfrost_end_query(struct pipe_context *pipe, struct pipe_query *q)
       ctx->occlusion_query = NULL;
       ctx->dirty |= PAN_DIRTY_OQ;
       break;
+   case PIPE_QUERY_TIMESTAMP: {
+      struct panfrost_resource *rsrc = pan_resource(query->rsrc);
+      panfrost_emit_write_timestamp(ctx, rsrc, 0, "TIMESTAMP end_query");
+      break;
+   }
+   case PIPE_QUERY_TIMESTAMP_DISJOINT:
+      /* This is a no op */
+      break;
+   case PIPE_QUERY_TIME_ELAPSED: {
+      struct panfrost_resource *rsrc = pan_resource(query->rsrc);
+      panfrost_emit_write_timestamp(ctx, rsrc, sizeof(uint64_t),
+                                    "TIME_ELAPSED end_query");
+      break;
+   }
    case PIPE_QUERY_PRIMITIVES_GENERATED:
       query->end = ctx->prims_generated;
       break;
@@ -716,6 +795,31 @@ panfrost_get_query_result(struct pipe_context *pipe, struct pipe_query *q,
 
       break;
 
+   case PIPE_QUERY_TIMESTAMP: {
+      panfrost_flush_writer(ctx, rsrc, "Timestamp query");
+      panfrost_bo_wait(rsrc->bo, INT64_MAX, false);
+      uint64_t *timestamp = (uint64_t *)rsrc->bo->ptr.cpu;
+
+      vresult->u64 = pan_gpu_time_to_ns(dev, *timestamp);
+      break;
+   }
+
+   case PIPE_QUERY_TIMESTAMP_DISJOINT: {
+      vresult->timestamp_disjoint.frequency =
+         dev->kmod.props.timestamp_frequency;
+      vresult->timestamp_disjoint.disjoint = false;
+      break;
+   }
+
+   case PIPE_QUERY_TIME_ELAPSED: {
+      panfrost_flush_writer(ctx, rsrc, "Time elapsed query");
+      panfrost_bo_wait(rsrc->bo, INT64_MAX, false);
+      uint64_t *timestamp = (uint64_t *)rsrc->bo->ptr.cpu;
+
+      vresult->u64 = pan_gpu_time_to_ns(dev, timestamp[1] - timestamp[0]);
+      break;
+   }
+
    case PIPE_QUERY_PRIMITIVES_GENERATED:
    case PIPE_QUERY_PRIMITIVES_EMITTED:
       panfrost_flush_all_batches(ctx, "Primitive count query");
@@ -794,7 +898,8 @@ static void
 panfrost_set_stream_output_targets(struct pipe_context *pctx,
                                    unsigned num_targets,
                                    struct pipe_stream_output_target **targets,
-                                   const unsigned *offsets)
+                                   const unsigned *offsets,
+                                   enum mesa_prim output_prim)
 {
    struct panfrost_context *ctx = pan_context(pctx);
    struct panfrost_streamout *so = &ctx->streamout;
@@ -897,6 +1002,8 @@ panfrost_create_context(struct pipe_screen *screen, void *priv, unsigned flags)
    if (!ctx)
       return NULL;
 
+   ctx->flags = flags;
+
    struct pipe_context *gallium = (struct pipe_context *)ctx;
    struct panfrost_device *dev = pan_device(screen);
 
@@ -987,11 +1094,13 @@ panfrost_create_context(struct pipe_screen *screen, void *priv, unsigned flags)
    gallium->stream_uploader = u_upload_create_default(gallium);
    gallium->const_uploader = gallium->stream_uploader;
 
-   panfrost_pool_init(&ctx->descs, ctx, dev, 0, 4096, "Descriptors", true,
-                      false);
+   if (panfrost_pool_init(&ctx->descs, ctx, dev, 0, 4096, "Descriptors", true,
+                          false) ||
 
-   panfrost_pool_init(&ctx->shaders, ctx, dev, PAN_BO_EXECUTE, 4096, "Shaders",
-                      true, false);
+       panfrost_pool_init(&ctx->shaders, ctx, dev, PAN_BO_EXECUTE, 4096,
+                          "Shaders", true, false)) {
+      goto failed;
+   }
 
    ctx->blitter = util_blitter_create(gallium);
 
@@ -1015,12 +1124,14 @@ panfrost_create_context(struct pipe_screen *screen, void *priv, unsigned flags)
 
    ret = pan_screen(screen)->vtbl.context_init(ctx);
 
-   if (ret) {
-      gallium->destroy(gallium);
-      return NULL;
-   }
+   if (ret)
+      goto failed;
 
    return gallium;
+
+failed:
+   gallium->destroy(gallium);
+   return NULL;
 }
 
 void
