@@ -26,11 +26,13 @@
 #include "drm-uapi/panthor_drm.h"
 
 #include "genxml/cs_builder.h"
+#include "panfrost/lib/genxml/cs_builder.h"
 
-#include "pan_blitter.h"
+#include "gen_macros.h"
 #include "pan_cmdstream.h"
 #include "pan_context.h"
 #include "pan_csf.h"
+#include "pan_fb_preload.h"
 #include "pan_job.h"
 
 #if PAN_ARCH < 10
@@ -55,30 +57,203 @@ csf_alloc_cs_buffer(void *cookie)
    };
 }
 
+/*
+ * Register is reserved to pass the batch tiler OOM context
+ */
+#define TILER_OOM_CTX_REG 76
+
+static enum cs_reg_perm
+csf_reg_perm_cb(struct cs_builder *b, unsigned reg)
+{
+   if (reg == TILER_OOM_CTX_REG)
+      return CS_REG_RD;
+   return CS_REG_RW;
+}
+
+static void
+csf_update_tiler_oom_ctx(struct cs_builder *b, uint64_t addr)
+{
+   reg_perm_cb_t orig_cb;
+
+   if (likely(!b->conf.reg_perm)) {
+      cs_move64_to(b, cs_reg64(b, TILER_OOM_CTX_REG), addr);
+      return;
+   }
+
+   orig_cb = b->conf.reg_perm;
+   b->conf.reg_perm = NULL;
+   cs_move64_to(b, cs_reg64(b, TILER_OOM_CTX_REG), addr);
+   b->conf.reg_perm = orig_cb;
+}
+
+#define FIELD_OFFSET(_name) offsetof(struct pan_csf_tiler_oom_ctx, _name)
+
+#define FBD_OFFSET(_pass)                                                      \
+   (FIELD_OFFSET(fbds) +                                                       \
+    (PAN_INCREMENTAL_RENDERING_##_pass##_PASS * sizeof(struct panfrost_ptr)) + \
+    offsetof(struct panfrost_ptr, gpu))
+
+static int
+csf_oom_handler_init(struct panfrost_context *ctx)
+{
+   struct panfrost_bo *cs_bo = NULL, *reg_save_bo = NULL;
+   struct panfrost_device *dev = pan_device(ctx->base.screen);
+
+   cs_bo =
+      panfrost_bo_create(dev, 4096, 0, "Temporary CS buffer");
+   reg_save_bo =
+      panfrost_bo_create(dev, 4096, 0, "reg save bo");
+
+   if (!cs_bo || !reg_save_bo)
+      goto fail;
+
+   struct cs_buffer queue = {
+      .cpu = cs_bo->ptr.cpu,
+      .gpu = cs_bo->ptr.gpu,
+      .capacity = panfrost_bo_size(cs_bo) / sizeof(uint64_t),
+   };
+   struct cs_builder b;
+   const struct cs_builder_conf conf = {
+      .nr_registers = 96,
+      .nr_kernel_registers = 4,
+      .reg_perm = (dev->debug & PAN_DBG_CS) ? csf_reg_perm_cb : NULL,
+   };
+   cs_builder_init(&b, &conf, queue);
+
+   struct cs_exception_handler_ctx handler_ctx = {
+      .ctx_reg = cs_reg64(&b, TILER_OOM_CTX_REG),
+      .dump_addr_offset = offsetof(struct pan_csf_tiler_oom_ctx, dump_addr),
+      .ls_sb_slot = 0,
+   };
+   struct cs_exception_handler handler;
+
+   cs_exception_handler_def(&b, &handler, handler_ctx) {
+      struct cs_index tiler_oom_ctx = cs_reg64(&b, TILER_OOM_CTX_REG);
+      struct cs_index counter = cs_reg32(&b, 47);
+      struct cs_index zero = cs_reg64(&b, 48);
+      struct cs_index flush_id = cs_reg32(&b, 48);
+      struct cs_index tiler_ctx = cs_reg64(&b, 50);
+      struct cs_index completed_top = cs_reg64(&b, 52);
+      struct cs_index completed_bottom = cs_reg64(&b, 54);
+      struct cs_index completed_chunks = cs_reg_tuple(&b, 52, 4);
+
+      /* Use different framebuffer descriptor depending on whether incremental
+       * rendering has already been triggered */
+      cs_load32_to(&b, counter, tiler_oom_ctx, FIELD_OFFSET(counter));
+      cs_wait_slot(&b, 0, false);
+      cs_if(&b, MALI_CS_CONDITION_GREATER, counter) {
+         cs_load64_to(&b, cs_reg64(&b, 40), tiler_oom_ctx, FBD_OFFSET(MIDDLE));
+      }
+      cs_else(&b) {
+         cs_load64_to(&b, cs_reg64(&b, 40), tiler_oom_ctx, FBD_OFFSET(FIRST));
+      }
+
+      cs_load32_to(&b, cs_reg32(&b, 42), tiler_oom_ctx, FIELD_OFFSET(bbox_min));
+      cs_load32_to(&b, cs_reg32(&b, 43), tiler_oom_ctx, FIELD_OFFSET(bbox_max));
+      cs_move64_to(&b, cs_reg64(&b, 44), 0);
+      cs_move32_to(&b, cs_reg32(&b, 46), 0);
+      cs_wait_slot(&b, 0, false);
+
+      /* Run the fragment job and wait */
+      cs_set_scoreboard_entry(&b, 3, 0);
+      cs_run_fragment(&b, false, MALI_TILE_RENDER_ORDER_Z_ORDER, false);
+      cs_wait_slot(&b, 3, false);
+
+      /* Increment counter */
+      cs_add32(&b, counter, counter, 1);
+      cs_store32(&b, counter, tiler_oom_ctx, FIELD_OFFSET(counter));
+
+      /* Load completed chunks */
+      cs_load64_to(&b, tiler_ctx, tiler_oom_ctx, FIELD_OFFSET(tiler_desc));
+      cs_wait_slot(&b, 0, false);
+      cs_load_to(&b, completed_chunks, tiler_ctx, BITFIELD_MASK(4), 10 * 4);
+      cs_wait_slot(&b, 0, false);
+
+      cs_finish_fragment(&b, false, completed_top, completed_bottom, cs_now());
+
+      /* Zero out polygon list, completed_top and completed_bottom */
+      cs_move64_to(&b, zero, 0);
+      cs_store64(&b, zero, tiler_ctx, 0);
+      cs_store64(&b, zero, tiler_ctx, 10 * 4);
+      cs_store64(&b, zero, tiler_ctx, 12 * 4);
+
+      /* We need to flush the texture caches so future preloads see the new
+       * content. */
+      cs_flush_caches(&b, MALI_CS_FLUSH_MODE_NONE, MALI_CS_FLUSH_MODE_NONE,
+                      true, flush_id, cs_defer(0, 0));
+
+      cs_wait_slot(&b, 0, false);
+
+      cs_set_scoreboard_entry(&b, 2, 0);
+   }
+
+   assert(cs_is_valid(&b));
+   cs_finish(&b);
+   ctx->csf.tiler_oom_handler.cs_bo = cs_bo;
+   ctx->csf.tiler_oom_handler.length = handler.length * sizeof(uint64_t);
+   ctx->csf.tiler_oom_handler.save_bo = reg_save_bo;
+
+   return 0;
+
+fail:
+   if (cs_bo)
+      panfrost_bo_unreference(cs_bo);
+
+   if (reg_save_bo)
+      panfrost_bo_unreference(reg_save_bo);
+
+   return -1;
+}
+
+#undef FBD_OFFSET
+#undef FIELD_OFFSET
+
 void
 GENX(csf_cleanup_batch)(struct panfrost_batch *batch)
 {
    free(batch->csf.cs.builder);
+   free(batch->csf.cs.ls_tracker);
 
    panfrost_pool_cleanup(&batch->csf.cs_chunk_pool);
 }
 
-void
+static inline struct panfrost_ptr
+alloc_fbd(struct panfrost_batch *batch)
+{
+   return pan_pool_alloc_desc_aggregate(
+      &batch->pool.base, PAN_DESC(FRAMEBUFFER), PAN_DESC(ZS_CRC_EXTENSION),
+      PAN_DESC_ARRAY(MAX2(batch->key.nr_cbufs, 1), RENDER_TARGET));
+}
+
+int
 GENX(csf_init_batch)(struct panfrost_batch *batch)
 {
    struct panfrost_device *dev = pan_device(batch->ctx->base.screen);
 
    /* Initialize the CS chunk pool. */
-   panfrost_pool_init(&batch->csf.cs_chunk_pool, NULL, dev, 0, 32768,
-                      "CS chunk pool", false, true);
+   if (panfrost_pool_init(&batch->csf.cs_chunk_pool, NULL, dev, 0, 32768,
+                          "CS chunk pool", false, true))
+      return -1;
+
+   if (dev->debug & PAN_DBG_CS) {
+      /* Load/store tracker if extra checks are enabled. */
+      batch->csf.cs.ls_tracker =
+         calloc(1, sizeof(struct cs_load_store_tracker));
+      batch->csf.cs.ls_tracker->sb_slot = 0;
+   }
 
    /* Allocate and bind the command queue */
    struct cs_buffer queue = csf_alloc_cs_buffer(batch);
+   if (!queue.gpu)
+      return -1;
+
    const struct cs_builder_conf conf = {
       .nr_registers = 96,
       .nr_kernel_registers = 4,
       .alloc_buffer = csf_alloc_cs_buffer,
       .cookie = batch,
+      .ls_tracker = batch->csf.cs.ls_tracker,
+      .reg_perm = (dev->debug & PAN_DBG_CS) ? csf_reg_perm_cb : NULL,
    };
 
    /* Setup the queue builder */
@@ -91,10 +266,15 @@ GENX(csf_init_batch)(struct panfrost_batch *batch)
    struct cs_builder *b = batch->csf.cs.builder;
    cs_set_scoreboard_entry(b, 2, 0);
 
-   batch->framebuffer = pan_pool_alloc_desc_aggregate(
-      &batch->pool.base, PAN_DESC(FRAMEBUFFER), PAN_DESC(ZS_CRC_EXTENSION),
-      PAN_DESC_ARRAY(MAX2(batch->key.nr_cbufs, 1), RENDER_TARGET));
+   batch->framebuffer = alloc_fbd(batch);
+   if (!batch->framebuffer.gpu)
+      return -1;
+
    batch->tls = pan_pool_alloc_desc(&batch->pool.base, LOCAL_STORAGE);
+   if (!batch->tls.cpu)
+      return -1;
+
+   return 0;
 }
 
 static void
@@ -144,7 +324,7 @@ csf_submit_gsubmit(struct panfrost_context *ctx,
    return 0;
 }
 
-static void
+static int
 csf_emit_batch_end(struct panfrost_batch *batch)
 {
    struct panfrost_device *dev = pan_device(batch->ctx->base.screen);
@@ -156,6 +336,9 @@ csf_emit_batch_end(struct panfrost_batch *batch)
    if (dev->debug & PAN_DBG_SYNC) {
       /* Get the CS state */
       batch->csf.cs.state = pan_pool_alloc_aligned(&batch->pool.base, 8, 8);
+      if (!batch->csf.cs.state.cpu)
+         return -1;
+
       memset(batch->csf.cs.state.cpu, ~0, 8);
       cs_move64_to(b, cs_reg64(b, 90), batch->csf.cs.state.gpu);
       cs_store_state(b, cs_reg64(b, 90), 0, MALI_CS_STATE_ERROR_STATUS,
@@ -166,12 +349,15 @@ csf_emit_batch_end(struct panfrost_batch *batch)
    struct cs_index flush_id = cs_reg32(b, 74);
    cs_move32_to(b, flush_id, 0);
    cs_flush_caches(b, MALI_CS_FLUSH_MODE_CLEAN, MALI_CS_FLUSH_MODE_CLEAN, true,
-                   flush_id, cs_now());
+                   flush_id, cs_defer(0, 0));
    cs_wait_slot(b, 0, false);
 
    /* Finish the command stream */
-   assert(cs_is_valid(batch->csf.cs.builder));
+   if (!cs_is_valid(batch->csf.cs.builder))
+      return -1;
+
    cs_finish(batch->csf.cs.builder);
+   return 0;
 }
 
 static int
@@ -368,6 +554,13 @@ csf_submit_wait_and_dump(struct panfrost_batch *batch,
          drmSyncobjTimelineWait(panfrost_device_fd(dev), &vm_sync_handle,
                                 &vm_sync_signal_point, 1, INT64_MAX, 0, NULL);
       assert(ret >= 0);
+
+      struct pan_csf_tiler_oom_ctx *tiler_oom_ctx =
+         batch->csf.tiler_oom_ctx.cpu;
+      if (tiler_oom_ctx != NULL && tiler_oom_ctx->counter > 0) {
+         perf_debug(ctx, "Incremental rendering was triggered %i time(s)",
+                    tiler_oom_ctx->counter);
+      }
    }
 
    /* Jobs won't be complete if blackhole rendering, that's ok */
@@ -383,9 +576,9 @@ csf_submit_wait_and_dump(struct panfrost_batch *batch,
 
       for (unsigned i = 0; i < gsubmit->queue_submits.count; i++) {
          uint32_t regs[256] = {0};
-         pandecode_cs(dev->decode_ctx, qsubmits[i].stream_addr,
-                      qsubmits[i].stream_size, panfrost_device_gpu_id(dev),
-                      regs);
+         pandecode_interpret_cs(dev->decode_ctx, qsubmits[i].stream_addr,
+                                qsubmits[i].stream_size,
+                                panfrost_device_gpu_id(dev), regs);
       }
 
       if (dev->debug & PAN_DBG_DUMP)
@@ -393,8 +586,7 @@ csf_submit_wait_and_dump(struct panfrost_batch *batch,
    }
 
    if (crash) {
-      fprintf(stderr, "Incomplete job or timeout\n");
-      fflush(NULL);
+      mesa_loge("Incomplete job or timeout\n");
       abort();
    }
 }
@@ -402,17 +594,19 @@ csf_submit_wait_and_dump(struct panfrost_batch *batch,
 int
 GENX(csf_submit_batch)(struct panfrost_batch *batch)
 {
-   /* Close the batch before submitting. */
-   csf_emit_batch_end(batch);
+   int ret;
 
-   uint32_t cs_instr_count = batch->csf.cs.builder->root_chunk.size;
-   uint64_t cs_start = batch->csf.cs.builder->root_chunk.buffer.gpu;
-   uint32_t cs_size = cs_instr_count * 8;
+   /* Close the batch before submitting. */
+   ret = csf_emit_batch_end(batch);
+   if (ret)
+      return ret;
+
+   uint64_t cs_start = cs_root_chunk_gpu_addr(batch->csf.cs.builder);
+   uint32_t cs_size = cs_root_chunk_size(batch->csf.cs.builder);
    struct panfrost_context *ctx = batch->ctx;
    struct panfrost_device *dev = pan_device(ctx->base.screen);
    uint32_t vm_sync_handle = panthor_kmod_vm_sync_handle(dev->kmod.vm);
    struct util_dynarray syncops;
-   int ret;
 
    util_dynarray_init(&syncops, NULL);
 
@@ -456,13 +650,148 @@ out_free_syncops:
    return ret;
 }
 
+static uint64_t
+csf_get_tiler_desc(struct panfrost_batch *batch)
+{
+   if (batch->tiler_ctx.valhall.desc)
+      return batch->tiler_ctx.valhall.desc;
+
+   struct panfrost_ptr t =
+      pan_pool_alloc_desc(&batch->pool.base, TILER_CONTEXT);
+
+   batch->csf.pending_tiler_desc = t.cpu;
+   batch->tiler_ctx.valhall.desc = t.gpu;
+   return batch->tiler_ctx.valhall.desc;
+}
+
+static void
+csf_emit_tiler_desc(struct panfrost_batch *batch, const struct pan_fb_info *fb)
+{
+   struct panfrost_context *ctx = batch->ctx;
+   struct panfrost_device *dev = pan_device(ctx->base.screen);
+
+   if (!batch->csf.pending_tiler_desc)
+      return;
+
+   pan_pack(batch->csf.pending_tiler_desc, TILER_CONTEXT, tiler) {
+      tiler.hierarchy_mask =
+         pan_select_tiler_hierarchy_mask(batch->key.width,
+                                         batch->key.height,
+                                         dev->tiler_features.max_levels);
+
+      /* For effective tile size larger than 16x16, disable first level */
+      if (fb->tile_size > 16 * 16)
+         tiler.hierarchy_mask &= ~1;
+
+      tiler.fb_width = batch->key.width;
+      tiler.fb_height = batch->key.height;
+      tiler.heap = batch->ctx->csf.heap.desc_bo->ptr.gpu;
+      tiler.sample_pattern =
+         pan_sample_pattern(util_framebuffer_get_num_samples(&batch->key));
+      tiler.first_provoking_vertex =
+         batch->first_provoking_vertex == U_TRISTATE_YES;
+      tiler.geometry_buffer = ctx->csf.tmp_geom_bo->ptr.gpu;
+      tiler.geometry_buffer_size = ctx->csf.tmp_geom_bo->kmod_bo->size;
+   }
+
+   batch->csf.pending_tiler_desc = NULL;
+}
+
+void
+GENX(csf_prepare_tiler)(struct panfrost_batch *batch, struct pan_fb_info *fb)
+{
+   csf_emit_tiler_desc(batch, fb);
+}
+
 void
 GENX(csf_preload_fb)(struct panfrost_batch *batch, struct pan_fb_info *fb)
 {
    struct panfrost_device *dev = pan_device(batch->ctx->base.screen);
 
    GENX(pan_preload_fb)
-   (&dev->blitter, &batch->pool.base, fb, 0, batch->tls.gpu, NULL);
+   (&dev->fb_preload_cache, &batch->pool.base, fb, batch->tls.gpu, NULL);
+}
+
+#define GET_FBD(_ctx, _pass)                                                   \
+   (_ctx)->fbds[PAN_INCREMENTAL_RENDERING_##_pass##_PASS]
+#define EMIT_FBD(_ctx, _pass, _fb, _tls, _tiler_ctx)                           \
+   GET_FBD(_ctx, _pass).gpu |=                                                 \
+      GENX(pan_emit_fbd)(_fb, 0, _tls, _tiler_ctx, GET_FBD(_ctx, _pass).cpu)
+
+void
+GENX(csf_emit_fbds)(struct panfrost_batch *batch, struct pan_fb_info *fb,
+                    struct pan_tls_info *tls)
+{
+   struct panfrost_device *dev = pan_device(batch->ctx->base.screen);
+
+   /* Default framebuffer descriptor */
+
+   batch->framebuffer.gpu |=
+      GENX(pan_emit_fbd)(fb, 0, tls, &batch->tiler_ctx, batch->framebuffer.cpu);
+
+   if (batch->draw_count == 0)
+      return;
+
+   struct pan_csf_tiler_oom_ctx *tiler_oom_ctx = batch->csf.tiler_oom_ctx.cpu;
+   struct pan_fb_info alt_fb;
+   bool changed = false;
+
+   /* First incremental rendering pass: don't discard result */
+
+   memcpy(&alt_fb, fb, sizeof(alt_fb));
+   for (unsigned i = 0; i < fb->rt_count; i++)
+      alt_fb.rts[i].discard = false;
+   alt_fb.zs.discard.z = false;
+   alt_fb.zs.discard.s = false;
+
+   EMIT_FBD(tiler_oom_ctx, FIRST, &alt_fb, tls, &batch->tiler_ctx);
+
+   /* Subsequent incremental rendering passes: preload old content and don't
+    * discard result */
+
+   for (unsigned i = 0; i < fb->rt_count; i++) {
+      if (fb->rts[i].view && !fb->rts[i].preload) {
+         alt_fb.rts[i].preload = true;
+         changed = true;
+      }
+
+      if (alt_fb.rts[i].clear) {
+         alt_fb.rts[i].clear = false;
+         changed = true;
+      }
+   }
+   if (fb->zs.view.zs && !fb->zs.preload.z && !fb->zs.preload.s) {
+      alt_fb.zs.preload.z = true;
+      alt_fb.zs.preload.s = true;
+      changed = true;
+   } else if (fb->zs.view.s && !fb->zs.preload.s) {
+      alt_fb.zs.preload.s = true;
+      changed = true;
+   }
+
+   if (alt_fb.zs.clear.z || alt_fb.zs.clear.s) {
+      alt_fb.zs.clear.z = false;
+      alt_fb.zs.clear.s = false;
+      changed = true;
+   }
+
+   if (changed) {
+      alt_fb.bifrost.pre_post.dcds.gpu = 0;
+      GENX(pan_preload_fb)
+      (&dev->fb_preload_cache, &batch->pool.base, &alt_fb, batch->tls.gpu, NULL);
+   }
+
+   EMIT_FBD(tiler_oom_ctx, MIDDLE, &alt_fb, tls, &batch->tiler_ctx);
+
+   /* Last incremental rendering pass: preload previous content and deal with
+    * results as specified by user */
+
+   for (unsigned i = 0; i < fb->rt_count; i++)
+      alt_fb.rts[i].discard = fb->rts[i].discard;
+   alt_fb.zs.discard.z = fb->zs.discard.z;
+   alt_fb.zs.discard.s = fb->zs.discard.s;
+
+   EMIT_FBD(tiler_oom_ctx, LAST, &alt_fb, tls, &batch->tiler_ctx);
 }
 
 void
@@ -470,6 +799,7 @@ GENX(csf_emit_fragment_job)(struct panfrost_batch *batch,
                             const struct pan_fb_info *pfb)
 {
    struct cs_builder *b = batch->csf.cs.builder;
+   struct pan_csf_tiler_oom_ctx *oom_ctx = batch->csf.tiler_oom_ctx.cpu;
 
    if (batch->draw_count > 0) {
       /* Finish tiling and wait for IDVS and tiling */
@@ -483,6 +813,19 @@ GENX(csf_emit_fragment_job)(struct panfrost_batch *batch,
    cs_move32_to(b, cs_reg32(b, 42), (batch->miny << 16) | batch->minx);
    cs_move32_to(b, cs_reg32(b, 43),
                 ((batch->maxy - 1) << 16) | (batch->maxx - 1));
+   cs_move64_to(b, cs_reg64(b, 44), 0);
+   cs_move32_to(b, cs_reg32(b, 46), 0);
+
+   /* Use different framebuffer descriptor if incremental rendering was
+    * triggered while tiling */
+   if (batch->draw_count > 0) {
+      struct cs_index counter = cs_reg32(b, 78);
+      cs_load32_to(b, counter, cs_reg64(b, TILER_OOM_CTX_REG), 0);
+      cs_wait_slot(b, 0, false);
+      cs_if(b, MALI_CS_CONDITION_GREATER, counter) {
+         cs_move64_to(b, cs_reg64(b, 40), GET_FBD(oom_ctx, LAST).gpu);
+      }
+   }
 
    /* Run the fragment job and wait */
    cs_run_fragment(b, false, MALI_TILE_RENDER_ORDER_Z_ORDER, false);
@@ -494,8 +837,8 @@ GENX(csf_emit_fragment_job)(struct panfrost_batch *batch,
     * chunks is in the tiler context descriptor
     * (completed_{top,bottom fields}). */
    if (batch->draw_count > 0) {
-      assert(batch->tiler_ctx.bifrost);
-      cs_move64_to(b, cs_reg64(b, 90), batch->tiler_ctx.bifrost);
+      assert(batch->tiler_ctx.valhall.desc);
+      cs_move64_to(b, cs_reg64(b, 90), batch->tiler_ctx.valhall.desc);
       cs_load_to(b, cs_reg_tuple(b, 86, 4), cs_reg64(b, 90), BITFIELD_MASK(4),
                  40);
       cs_wait_slot(b, 0, false);
@@ -505,9 +848,9 @@ GENX(csf_emit_fragment_job)(struct panfrost_batch *batch,
 
 static void
 csf_emit_shader_regs(struct panfrost_batch *batch, enum pipe_shader_type stage,
-                     mali_ptr shader)
+                     uint64_t shader)
 {
-   mali_ptr resources = panfrost_emit_resources(batch, stage);
+   uint64_t resources = panfrost_emit_resources(batch, stage);
 
    assert(stage == PIPE_SHADER_VERTEX || stage == PIPE_SHADER_FRAGMENT ||
           stage == PIPE_SHADER_COMPUTE);
@@ -544,8 +887,8 @@ GENX(csf_launch_grid)(struct panfrost_batch *batch,
    cs_move32_to(b, cs_reg32(b, 32), 0);
 
    /* Compute workgroup size */
-   uint32_t wg_size[4];
-   pan_pack(wg_size, COMPUTE_SIZE_WORKGROUP, cfg) {
+   struct mali_compute_size_workgroup_packed wg_size;
+   pan_pack(&wg_size, COMPUTE_SIZE_WORKGROUP, cfg) {
       cfg.workgroup_size_x = info->block[0];
       cfg.workgroup_size_y = info->block[1];
       cfg.workgroup_size_z = info->block[2];
@@ -560,7 +903,7 @@ GENX(csf_launch_grid)(struct panfrost_batch *batch,
                                      (info->variable_shared_mem == 0);
    }
 
-   cs_move32_to(b, cs_reg32(b, 33), wg_size[0]);
+   cs_move32_to(b, cs_reg32(b, 33), wg_size.opaque[0]);
 
    /* Offset */
    for (unsigned i = 0; i < 3; ++i)
@@ -647,8 +990,8 @@ GENX(csf_launch_xfb)(struct panfrost_batch *batch,
    cs_move32_to(b, cs_reg32(b, 32), batch->ctx->offset_start);
 
    /* Compute workgroup size */
-   uint32_t wg_size[4];
-   pan_pack(wg_size, COMPUTE_SIZE_WORKGROUP, cfg) {
+   struct mali_compute_size_workgroup_packed wg_size;
+   pan_pack(&wg_size, COMPUTE_SIZE_WORKGROUP, cfg) {
       cfg.workgroup_size_x = 1;
       cfg.workgroup_size_y = 1;
       cfg.workgroup_size_z = 1;
@@ -658,7 +1001,7 @@ GENX(csf_launch_xfb)(struct panfrost_batch *batch,
        */
       cfg.allow_merging_workgroups = true;
    }
-   cs_move32_to(b, cs_reg32(b, 33), wg_size[0]);
+   cs_move32_to(b, cs_reg32(b, 33), wg_size.opaque[0]);
 
    /* Offset */
    for (unsigned i = 0; i < 3; ++i)
@@ -677,56 +1020,35 @@ GENX(csf_launch_xfb)(struct panfrost_batch *batch,
    cs_run_compute(b, 1, MALI_TASK_AXIS_Z, false, cs_shader_res_sel(0, 0, 0, 0));
 }
 
-static mali_ptr
-csf_get_tiler_desc(struct panfrost_batch *batch)
+static void
+emit_tiler_oom_context(struct cs_builder *b, struct panfrost_batch *batch)
 {
-   struct panfrost_context *ctx = batch->ctx;
-   struct panfrost_device *dev = pan_device(ctx->base.screen);
+   struct pan_csf_tiler_oom_ctx *ctx;
 
-   if (batch->tiler_ctx.bifrost)
-      return batch->tiler_ctx.bifrost;
+   batch->csf.tiler_oom_ctx =
+      pan_pool_alloc_aligned(&batch->pool.base, sizeof(*ctx), 8);
+   ctx = batch->csf.tiler_oom_ctx.cpu;
 
-   struct panfrost_ptr t =
-      pan_pool_alloc_desc(&batch->pool.base, TILER_CONTEXT);
-   pan_pack(t.cpu, TILER_CONTEXT, tiler) {
-      unsigned max_levels = dev->tiler_features.max_levels;
-      assert(max_levels >= 2);
+   ctx->tiler_desc = csf_get_tiler_desc(batch);
+   ctx->counter = 0;
+   ctx->bbox_min = (batch->miny << 16) | batch->minx;
+   ctx->bbox_max = ((batch->maxy - 1) << 16) | (batch->maxx - 1);
+   ctx->dump_addr = batch->ctx->csf.tiler_oom_handler.save_bo->ptr.gpu;
 
-      /* TODO: Select hierarchy mask more effectively */
-      tiler.hierarchy_mask = (max_levels >= 8) ? 0xFF : 0x28;
+   for (unsigned i = 0; i < PAN_INCREMENTAL_RENDERING_PASS_COUNT; ++i)
+      ctx->fbds[i] = alloc_fbd(batch);
 
-      /* For large framebuffers, disable the smallest bin size to
-       * avoid pathological tiler memory usage. Required to avoid OOM
-       * on dEQP-GLES31.functional.fbo.no_attachments.maximums.all on
-       * Mali-G57.
-       */
-      if (MAX2(batch->key.width, batch->key.height) >= 4096)
-         tiler.hierarchy_mask &= ~1;
-
-      tiler.fb_width = batch->key.width;
-      tiler.fb_height = batch->key.height;
-      tiler.heap = batch->ctx->csf.heap.desc_bo->ptr.gpu;
-      tiler.sample_pattern =
-         pan_sample_pattern(util_framebuffer_get_num_samples(&batch->key));
-      tiler.first_provoking_vertex =
-         pan_tristate_get(batch->first_provoking_vertex);
-      tiler.geometry_buffer = ctx->csf.tmp_geom_bo->ptr.gpu;
-      tiler.geometry_buffer_size = ctx->csf.tmp_geom_bo->kmod_bo->size;
-   }
-
-   batch->tiler_ctx.bifrost = t.gpu;
-   return batch->tiler_ctx.bifrost;
+   csf_update_tiler_oom_ctx(b, batch->csf.tiler_oom_ctx.gpu);
 }
 
-void
-GENX(csf_launch_draw)(struct panfrost_batch *batch,
-                      const struct pipe_draw_info *info, unsigned drawid_offset,
-                      const struct pipe_draw_start_count_bias *draw,
-                      unsigned vertex_count)
+static uint32_t
+csf_emit_draw_state(struct panfrost_batch *batch,
+                    const struct pipe_draw_info *info, unsigned drawid_offset)
 {
    struct panfrost_context *ctx = batch->ctx;
    struct panfrost_compiled_shader *vs = ctx->prog[PIPE_SHADER_VERTEX];
    struct panfrost_compiled_shader *fs = ctx->prog[PIPE_SHADER_FRAGMENT];
+
    bool idvs = vs->info.vs.idvs;
    bool fs_required = panfrost_fs_required(
       fs, ctx->blend, &ctx->pipe_framebuffer, ctx->depth_stencil);
@@ -736,8 +1058,10 @@ GENX(csf_launch_draw)(struct panfrost_batch *batch,
 
    struct cs_builder *b = batch->csf.cs.builder;
 
-   if (batch->draw_count == 0)
+   if (batch->draw_count == 0) {
+      emit_tiler_oom_context(b, batch);
       cs_vt_start(batch->csf.cs.builder, cs_now());
+   }
 
    csf_emit_shader_regs(batch, PIPE_SHADER_VERTEX,
                         panfrost_get_position_shader(batch, info));
@@ -758,20 +1082,6 @@ GENX(csf_launch_draw)(struct panfrost_batch *batch,
    cs_move64_to(b, cs_reg64(b, 24), batch->tls.gpu);
    cs_move64_to(b, cs_reg64(b, 30), batch->tls.gpu);
    cs_move32_to(b, cs_reg32(b, 32), 0);
-   cs_move32_to(b, cs_reg32(b, 33), draw->count);
-   cs_move32_to(b, cs_reg32(b, 34), info->instance_count);
-   cs_move32_to(b, cs_reg32(b, 35), 0);
-
-   /* Base vertex offset on Valhall is used for both indexed and
-    * non-indexed draws, in a simple way for either. Handle both cases.
-    */
-   if (info->index_size) {
-      cs_move32_to(b, cs_reg32(b, 36), draw->index_bias);
-      cs_move32_to(b, cs_reg32(b, 39), info->index_size * draw->count);
-   } else {
-      cs_move32_to(b, cs_reg32(b, 36), draw->start);
-      cs_move32_to(b, cs_reg32(b, 39), 0);
-   }
    cs_move32_to(b, cs_reg32(b, 37), 0);
    cs_move32_to(b, cs_reg32(b, 38), 0);
 
@@ -799,13 +1109,18 @@ GENX(csf_launch_draw)(struct panfrost_batch *batch,
    if (info->index_size)
       cs_move64_to(b, cs_reg64(b, 54), batch->indices);
 
-   uint32_t primitive_flags = 0;
+   struct pipe_rasterizer_state *rast = &ctx->rasterizer->base;
+
+   struct mali_primitive_flags_packed primitive_flags;
    pan_pack(&primitive_flags, PRIMITIVE_FLAGS, cfg) {
       if (panfrost_writes_point_size(ctx))
          cfg.point_size_array_format = MALI_POINT_SIZE_ARRAY_FORMAT_FP16;
 
       cfg.allow_rotating_primitives = allow_rotating_primitives(fs, info);
 
+      cfg.low_depth_cull = rast->depth_clip_near;
+      cfg.high_depth_cull = rast->depth_clip_far;
+
       /* Non-fixed restart indices should have been lowered */
       assert(!cfg.primitive_restart || panfrost_is_implicit_prim_restart(info));
       cfg.primitive_restart = info->primitive_restart;
@@ -815,11 +1130,11 @@ GENX(csf_launch_draw)(struct panfrost_batch *batch,
                                     : MALI_FIFO_FORMAT_BASIC;
    }
 
-   cs_move32_to(b, cs_reg32(b, 56), primitive_flags);
+   cs_move32_to(b, cs_reg32(b, 56), primitive_flags.opaque[0]);
 
-   struct pipe_rasterizer_state *rast = &ctx->rasterizer->base;
+   struct mali_dcd_flags_0_packed dcd_flags0;
+   struct mali_dcd_flags_1_packed dcd_flags1;
 
-   uint32_t dcd_flags0 = 0, dcd_flags1 = 0;
    pan_pack(&dcd_flags0, DCD_FLAGS_0, cfg) {
       enum mesa_prim reduced_mode = u_reduced_prim(info->mode);
       bool polygon = reduced_mode == MESA_PRIM_TRIANGLES;
@@ -881,7 +1196,8 @@ GENX(csf_launch_draw)(struct panfrost_batch *batch,
 
          /* Also use per-sample shading if required by the shader
           */
-         cfg.evaluate_per_sample |= fs->info.fs.sample_shading;
+         cfg.evaluate_per_sample |=
+            (fs->info.fs.sample_shading && rast->multisample);
 
          /* Unlike Bifrost, alpha-to-coverage must be included in
           * this identically-named flag. Confusing, isn't it?
@@ -924,27 +1240,132 @@ GENX(csf_launch_draw)(struct panfrost_batch *batch,
       }
    }
 
-   cs_move32_to(b, cs_reg32(b, 57), dcd_flags0);
-   cs_move32_to(b, cs_reg32(b, 58), dcd_flags1);
+   cs_move32_to(b, cs_reg32(b, 57), dcd_flags0.opaque[0]);
+   cs_move32_to(b, cs_reg32(b, 58), dcd_flags1.opaque[0]);
 
-   uint64_t primsize = 0;
+   struct mali_primitive_size_packed primsize;
    panfrost_emit_primitive_size(ctx, info->mode == MESA_PRIM_POINTS, 0,
                                 &primsize);
-   cs_move64_to(b, cs_reg64(b, 60), primsize);
+   struct mali_primitive_size_packed *primsize_ptr = &primsize;
+   cs_move64_to(b, cs_reg64(b, 60), *((uint64_t*)primsize_ptr));
 
-   uint32_t flags_override;
-   pan_pack(&flags_override, PRIMITIVE_FLAGS, cfg) {
+   struct mali_primitive_flags_packed flags_override;
+   /* Pack with nodefaults so only explicitly set override fields affect the
+    * previously set register values */
+   pan_pack_nodefaults(&flags_override, PRIMITIVE_FLAGS, cfg) {
       cfg.draw_mode = pan_draw_mode(info->mode);
       cfg.index_type = panfrost_translate_index_size(info->index_size);
       cfg.secondary_shader = secondary_shader;
+   };
+
+   return flags_override.opaque[0];
+}
+
+static struct cs_index
+csf_emit_draw_id_register(struct panfrost_batch *batch, unsigned offset)
+{
+   struct cs_builder *b = batch->csf.cs.builder;
+   struct panfrost_context *ctx = batch->ctx;
+   struct panfrost_uncompiled_shader *vs = ctx->uncompiled[PIPE_SHADER_VERTEX];
+
+   if (!BITSET_TEST(vs->nir->info.system_values_read, SYSTEM_VALUE_DRAW_ID))
+      return cs_undef();
+
+   struct cs_index drawid = cs_reg32(b, 67);
+   cs_move32_to(b, drawid, offset);
+
+   return drawid;
+}
+
+void
+GENX(csf_launch_draw)(struct panfrost_batch *batch,
+                      const struct pipe_draw_info *info, unsigned drawid_offset,
+                      const struct pipe_draw_start_count_bias *draw,
+                      unsigned vertex_count)
+{
+   struct cs_builder *b = batch->csf.cs.builder;
+
+   uint32_t flags_override = csf_emit_draw_state(batch, info, drawid_offset);
+   struct cs_index drawid = csf_emit_draw_id_register(batch, drawid_offset);
+
+   cs_move32_to(b, cs_reg32(b, 33), draw->count);
+   cs_move32_to(b, cs_reg32(b, 34), info->instance_count);
+   cs_move32_to(b, cs_reg32(b, 35), 0);
+
+   /* Base vertex offset on Valhall is used for both indexed and
+    * non-indexed draws, in a simple way for either. Handle both cases.
+    */
+   if (info->index_size) {
+      cs_move32_to(b, cs_reg32(b, 36), draw->index_bias);
+      cs_move32_to(b, cs_reg32(b, 39), info->index_size * draw->count);
+   } else {
+      cs_move32_to(b, cs_reg32(b, 36), draw->start);
+      cs_move32_to(b, cs_reg32(b, 39), 0);
    }
 
    cs_run_idvs(b, flags_override, false, true, cs_shader_res_sel(0, 0, 1, 0),
-               cs_shader_res_sel(2, 2, 2, 0), cs_undef());
+               cs_shader_res_sel(2, 2, 2, 0), drawid);
+}
+
+void
+GENX(csf_launch_draw_indirect)(struct panfrost_batch *batch,
+                               const struct pipe_draw_info *info,
+                               unsigned drawid_offset,
+                               const struct pipe_draw_indirect_info *indirect)
+{
+   struct cs_builder *b = batch->csf.cs.builder;
+
+   uint32_t flags_override = csf_emit_draw_state(batch, info, drawid_offset);
+   struct cs_index drawid = csf_emit_draw_id_register(batch, drawid_offset);
+
+   struct cs_index address = cs_reg64(b, 64);
+   struct cs_index counter = cs_reg32(b, 66);
+   cs_move64_to(
+      b, address,
+      pan_resource(indirect->buffer)->image.data.base + indirect->offset);
+   cs_move32_to(b, counter, indirect->draw_count);
+
+   cs_while(b, MALI_CS_CONDITION_GREATER, counter) {
+      if (info->index_size) {
+         /* loads vertex count, instance count, index offset, vertex offset */
+         cs_load_to(b, cs_reg_tuple(b, 33, 4), address, BITFIELD_MASK(4), 0);
+         cs_move32_to(b, cs_reg32(b, 39), info->index.resource->width0);
+      } else {
+         /* vertex count, instance count */
+         cs_load_to(b, cs_reg_tuple(b, 33, 2), address, BITFIELD_MASK(2), 0);
+         cs_move32_to(b, cs_reg32(b, 35), 0);
+         cs_load_to(b, cs_reg_tuple(b, 36, 1), address, BITFIELD_MASK(1),
+                    2 * sizeof(uint32_t)); // instance offset
+         cs_move32_to(b, cs_reg32(b, 37), 0);
+         cs_move32_to(b, cs_reg32(b, 39), 0);
+      }
+
+      cs_wait_slot(b, 0, false);
+      cs_run_idvs(b, flags_override, false, true, cs_shader_res_sel(0, 0, 1, 0),
+                  cs_shader_res_sel(2, 2, 2, 0), drawid);
+
+      cs_add64(b, address, address, indirect->stride);
+      cs_add32(b, counter, counter, (unsigned int)-1);
+      if (drawid.type != CS_INDEX_UNDEF)
+         cs_add32(b, drawid, drawid, 1);
+   }
 }
 
 #define POSITION_FIFO_SIZE (64 * 1024)
 
+static enum drm_panthor_group_priority
+get_panthor_group_priority(struct panfrost_context *ctx)
+{
+   if (ctx->flags & PIPE_CONTEXT_REALTIME_PRIORITY)
+      return PANTHOR_GROUP_PRIORITY_REALTIME;
+   else if (ctx->flags & PIPE_CONTEXT_HIGH_PRIORITY)
+      return PANTHOR_GROUP_PRIORITY_HIGH;
+   else if (ctx->flags & PIPE_CONTEXT_LOW_PRIORITY)
+      return PANTHOR_GROUP_PRIORITY_LOW;
+
+   return PANTHOR_GROUP_PRIORITY_MEDIUM;
+}
+
 int
 GENX(csf_init_context)(struct panfrost_context *ctx)
 {
@@ -961,7 +1382,7 @@ GENX(csf_init_context)(struct panfrost_context *ctx)
       .max_compute_cores = util_bitcount64(dev->kmod.props.shader_present),
       .max_fragment_cores = util_bitcount64(dev->kmod.props.shader_present),
       .max_tiler_cores = 1,
-      .priority = PANTHOR_GROUP_PRIORITY_MEDIUM,
+      .priority = get_panthor_group_priority(ctx),
       .queues = DRM_PANTHOR_OBJ_ARRAY(ARRAY_SIZE(qc), qc),
       .vm_id = pan_kmod_vm_handle(dev->kmod.vm),
    };
@@ -1005,7 +1426,7 @@ GENX(csf_init_context)(struct panfrost_context *ctx)
    if (ctx->csf.heap.desc_bo == NULL)
       goto err_tiler_heap_desc_bo;
 
-   pan_pack(ctx->csf.heap.desc_bo->ptr.cpu, TILER_HEAP, heap) {
+   pan_cast_and_pack(ctx->csf.heap.desc_bo->ptr.cpu, TILER_HEAP, heap) {
       heap.size = pan_screen(ctx->base.screen)->csf_tiler_heap.chunk_size;
       heap.base = thc.first_heap_chunk_gpu_va;
       heap.bottom = heap.base + 64;
@@ -1025,6 +1446,9 @@ GENX(csf_init_context)(struct panfrost_context *ctx)
    if (cs_bo == NULL)
       goto err_tiler_heap_cs_bo;
 
+   if (csf_oom_handler_init(ctx))
+      goto err_g_submit;
+
    struct cs_buffer init_buffer = {
       .cpu = cs_bo->ptr.cpu,
       .gpu = cs_bo->ptr.gpu,
@@ -1040,6 +1464,13 @@ GENX(csf_init_context)(struct panfrost_context *ctx)
    cs_move64_to(&b, heap, thc.tiler_heap_ctx_gpu_va);
    cs_heap_set(&b, heap);
 
+   struct cs_index addr_reg = cs_reg64(&b, 86);
+   struct cs_index length_reg = cs_reg32(&b, 88);
+   cs_move64_to(&b, addr_reg, ctx->csf.tiler_oom_handler.cs_bo->ptr.gpu);
+   cs_move32_to(&b, length_reg, ctx->csf.tiler_oom_handler.length);
+   cs_set_exception_handler(&b, MALI_CS_EXCEPTION_TYPE_TILER_OOM,
+                            addr_reg, length_reg);
+
    struct drm_panthor_queue_submit qsubmit;
    struct drm_panthor_group_submit gsubmit;
    struct drm_panthor_sync_op sync = {
@@ -1051,14 +1482,20 @@ GENX(csf_init_context)(struct panfrost_context *ctx)
    assert(cs_is_valid(&b));
    cs_finish(&b);
 
-   uint32_t cs_instr_count = b.root_chunk.size;
-   uint64_t cs_start = b.root_chunk.buffer.gpu;
-   uint32_t cs_size = cs_instr_count * 8;
+   uint64_t cs_start = cs_root_chunk_gpu_addr(&b);
+   uint32_t cs_size = cs_root_chunk_size(&b);
 
    csf_prepare_qsubmit(ctx, &qsubmit, 0, cs_start, cs_size, &sync, 1);
    csf_prepare_gsubmit(ctx, &gsubmit, &qsubmit, 1);
    ret = csf_submit_gsubmit(ctx, &gsubmit);
 
+   if (dev->debug & PAN_DBG_TRACE) {
+      uint32_t regs[256] = {0};
+      pandecode_interpret_cs(dev->decode_ctx, qsubmit.stream_addr,
+                             qsubmit.stream_size, panfrost_device_gpu_id(dev),
+                             regs);
+   }
+
    if (ret)
       goto err_g_submit;
 
@@ -1115,6 +1552,23 @@ GENX(csf_cleanup_context)(struct panfrost_context *ctx)
       drmIoctl(panfrost_device_fd(dev), DRM_IOCTL_PANTHOR_GROUP_DESTROY, &gd);
    assert(!ret);
 
+   panfrost_bo_unreference(ctx->csf.tmp_geom_bo);
    panfrost_bo_unreference(ctx->csf.heap.desc_bo);
+   panfrost_bo_unreference(ctx->csf.tiler_oom_handler.cs_bo);
+   panfrost_bo_unreference(ctx->csf.tiler_oom_handler.save_bo);
    ctx->csf.is_init = false;
 }
+
+void
+GENX(csf_emit_write_timestamp)(struct panfrost_batch *batch,
+                               struct panfrost_resource *dst, unsigned offset)
+{
+   struct cs_builder *b = batch->csf.cs.builder;
+
+   struct cs_index address = cs_reg64(b, 40);
+   cs_move64_to(b, address,
+                dst->image.data.base + dst->image.data.offset + offset);
+   cs_store_state(b, address, 0, MALI_CS_STATE_TIMESTAMP, cs_now());
+
+   panfrost_batch_write_rsrc(batch, dst, PIPE_SHADER_VERTEX);
+}
