@@ -70,7 +70,7 @@ panfrost_batch_add_surface(struct panfrost_batch *batch,
    }
 }
 
-static void
+static int
 panfrost_batch_init(struct panfrost_context *ctx,
                     const struct pipe_framebuffer_state *key,
                     struct panfrost_batch *batch)
@@ -92,21 +92,23 @@ panfrost_batch_init(struct panfrost_context *ctx,
 
    /* Preallocate the main pool, since every batch has at least one job
     * structure so it will be used */
-   panfrost_pool_init(&batch->pool, NULL, dev, 0, 65536, "Batch pool", true,
-                      true);
+   if (panfrost_pool_init(&batch->pool, NULL, dev, 0, 65536, "Batch pool",
+                          true, true))
+      return -1;
 
    /* Don't preallocate the invisible pool, since not every batch will use
     * the pre-allocation, particularly if the varyings are larger than the
     * preallocation and a reallocation is needed after anyway. */
-   panfrost_pool_init(&batch->invisible_pool, NULL, dev, PAN_BO_INVISIBLE,
-                      65536, "Varyings", false, true);
+   if (panfrost_pool_init(&batch->invisible_pool, NULL, dev,
+                          PAN_BO_INVISIBLE, 65536, "Varyings", false, true))
+      return -1;
 
    for (unsigned i = 0; i < batch->key.nr_cbufs; ++i)
       panfrost_batch_add_surface(batch, batch->key.cbufs[i]);
 
    panfrost_batch_add_surface(batch, batch->key.zsbuf);
 
-   screen->vtbl.init_batch(batch);
+   return screen->vtbl.init_batch(batch);
 }
 
 static void
@@ -184,7 +186,13 @@ panfrost_get_batch(struct panfrost_context *ctx,
       panfrost_batch_submit(ctx, batch);
    }
 
-   panfrost_batch_init(ctx, key, batch);
+   if (panfrost_batch_init(ctx, key, batch)) {
+      mesa_loge("panfrost_batch_init failed");
+      panfrost_batch_cleanup(ctx, batch);
+      /* prevent this batch from being reused without initializing */
+      batch->seqnum = 0;
+      return NULL;
+   }
 
    unsigned batch_idx = panfrost_batch_idx(batch);
    BITSET_SET(ctx->batches.active, batch_idx);
@@ -208,6 +216,8 @@ panfrost_get_batch_for_fbo(struct panfrost_context *ctx)
    /* If not, look up the job */
    struct panfrost_batch *batch =
       panfrost_get_batch(ctx, &ctx->pipe_framebuffer);
+   if (!batch)
+      return NULL;
 
    /* Set this job as the current FBO job. Will be reset when updating the
     * FB state and when submitting or releasing a job.
@@ -371,6 +381,8 @@ panfrost_batch_read_rsrc(struct panfrost_batch *batch,
 
    if (rsrc->separate_stencil)
       panfrost_batch_add_bo_old(batch, rsrc->separate_stencil->bo, access);
+   if (rsrc->shadow_image)
+      panfrost_batch_add_bo_old(batch, rsrc->shadow_image->bo, access);
 
    panfrost_batch_update_access(batch, rsrc, false);
 }
@@ -386,6 +398,8 @@ panfrost_batch_write_rsrc(struct panfrost_batch *batch,
 
    if (rsrc->separate_stencil)
       panfrost_batch_add_bo_old(batch, rsrc->separate_stencil->bo, access);
+   if (rsrc->shadow_image)
+      panfrost_batch_add_bo_old(batch, rsrc->shadow_image->bo, access);
 
    panfrost_batch_update_access(batch, rsrc, true);
 }
@@ -399,15 +413,16 @@ panfrost_batch_create_bo(struct panfrost_batch *batch, size_t size,
 
    bo = panfrost_bo_create(pan_device(batch->ctx->base.screen), size,
                            create_flags, label);
-   assert(bo);
-   panfrost_batch_add_bo(batch, bo, stage);
+   if (bo) {
+      panfrost_batch_add_bo(batch, bo, stage);
 
-   /* panfrost_batch_add_bo() has retained a reference and
-    * panfrost_bo_create() initialize the refcnt to 1, so let's
-    * unreference the BO here so it gets released when the batch is
-    * destroyed (unless it's retained by someone else in the meantime).
-    */
-   panfrost_bo_unreference(bo);
+      /* panfrost_batch_add_bo() has retained a reference and
+       * panfrost_bo_create() initialize the refcnt to 1, so let's
+       * unreference the BO here so it gets released when the batch is
+       * destroyed (unless it's retained by someone else in the meantime).
+       */
+      panfrost_bo_unreference(bo);
+   }
    return bo;
 }
 
@@ -426,7 +441,8 @@ panfrost_batch_get_scratchpad(struct panfrost_batch *batch,
          panfrost_batch_create_bo(batch, size, PAN_BO_INVISIBLE,
                                   PIPE_SHADER_VERTEX, "Thread local storage");
 
-      panfrost_batch_add_bo(batch, batch->scratchpad, PIPE_SHADER_FRAGMENT);
+      if (batch->scratchpad)
+         panfrost_batch_add_bo(batch, batch->scratchpad, PIPE_SHADER_FRAGMENT);
    }
 
    return batch->scratchpad;
@@ -454,6 +470,7 @@ panfrost_batch_to_fb_info(const struct panfrost_batch *batch,
                           bool reserve)
 {
    struct panfrost_device *dev = pan_device(batch->ctx->base.screen);
+   struct panfrost_screen *screen = pan_screen(batch->ctx->base.screen);
 
    memset(fb, 0, sizeof(*fb));
    memset(rts, 0, sizeof(*rts) * 8);
@@ -468,10 +485,11 @@ panfrost_batch_to_fb_info(const struct panfrost_batch *batch,
    fb->extent.maxx = batch->maxx - 1;
    fb->extent.maxy = batch->maxy - 1;
    fb->nr_samples = util_framebuffer_get_num_samples(&batch->key);
-   fb->force_samples = pan_tristate_get(batch->line_smoothing) ? 16 : 0;
+   fb->force_samples = (batch->line_smoothing == U_TRISTATE_YES) ? 16 : 0;
    fb->rt_count = batch->key.nr_cbufs;
-   fb->sprite_coord_origin = pan_tristate_get(batch->sprite_coord_origin);
-   fb->first_provoking_vertex = pan_tristate_get(batch->first_provoking_vertex);
+   fb->sprite_coord_origin = (batch->sprite_coord_origin == U_TRISTATE_YES);
+   fb->first_provoking_vertex =
+      (batch->first_provoking_vertex == U_TRISTATE_YES);
 
    static const unsigned char id_swz[] = {
       PIPE_SWIZZLE_X,
@@ -605,6 +623,8 @@ panfrost_batch_to_fb_info(const struct panfrost_batch *batch,
       fb->zs.preload.z = !fb->zs.clear.z && valid;
       fb->zs.preload.s = !fb->zs.clear.s && valid;
    }
+
+   screen->vtbl.select_tile_size(fb);
 }
 
 static void
@@ -634,7 +654,7 @@ panfrost_batch_submit(struct panfrost_context *ctx,
    int ret;
 
    /* Nothing to do! */
-   if (!has_frag && batch->compute_count == 0)
+   if (!has_frag && batch->compute_count == 0 && !batch->has_time_query)
       goto out;
 
    if (batch->key.zsbuf && has_frag) {
@@ -668,7 +688,7 @@ panfrost_batch_submit(struct panfrost_context *ctx,
 
    ret = screen->vtbl.submit_batch(batch, &fb);
    if (ret)
-      fprintf(stderr, "panfrost_batch_submit failed: %d\n", ret);
+      mesa_loge("panfrost_batch_submit failed: %d\n", ret);
 
    /* We must reset the damage info of our render targets here even
     * though a damage reset normally happens when the DRI layer swaps
@@ -700,6 +720,9 @@ panfrost_flush_all_batches(struct panfrost_context *ctx, const char *reason)
       perf_debug(ctx, "Flushing everything due to: %s", reason);
 
    struct panfrost_batch *batch = panfrost_get_batch_for_fbo(ctx);
+   if (!batch)
+      return;
+
    panfrost_batch_submit(ctx, batch);
 
    for (unsigned i = 0; i < PAN_MAX_BATCHES; i++) {
